from typing import Any, Dict, Union, Mapping, Iterable, Optional
import logging

import pandas as pd

from omnipath._core.query import QueryType
from omnipath._core.utils._docs import d
from omnipath._core.requests._request import OmnipathRequestABC
from omnipath.constants._pkg_constants import Key, final

_MAX_N_PROTS = 600


@final
class Annotations(OmnipathRequestABC):
    """Request annotations from [OmniPath]_."""

    __string__ = frozenset({"source", "value"})
    __categorical__ = frozenset({"entity_type", "label", "source"})

    _query_type = QueryType.ANNOTATIONS

    def _modify_params(self, params: Dict[str, Any]) -> Dict[str, Any]:
        params.pop(Key.ORGANISM.value, None)

        return params

    @classmethod
    @d.dedent
    def params(cls) -> Dict[str, Any]:
        """%(query_params)s"""
        params = super().params()
        params.pop(Key.ORGANISM.value, None)

        return params

    @classmethod
    def get(
        cls,
        proteins: Optional[Union[str, Iterable[str]]] = None,
        resources: Optional[Union[str, Iterable[str]]] = None,
        force_full_download: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Import annotations from [OmniPath]_.

        Retrieves protein annotations about function, localization, expression, structure and other properties of
        proteins from `OmniPath <https://omnipathdb.org/annotations>`__.

        Parameters
        ----------
        proteins
            Genes or proteins for which annotations will be retrieved (UniProt IDs, HGNC Gene Symbols or miRBase IDs).

            In order to download annotations for proteins complexes, write **'COMPLEX:'** before the gene symbols of
            the genes integrating the complex.

            If `None`, fetch annotations for all available genes or proteins.
        resources
            Load the annotations only from these databases. See :meth:`resources` for available options.
            If `None`, use all available resources.
        force_full_download
            Force the download of the entire annotations dataset. The full size of the data is ~1GB.
            We recommend to retrieve the annotations for a set of proteins or only from a few resources,
            depending on your interest.
        kwargs
            Additional query parameters.

        Returns
        -------
        :class:`pandas.DataFrame`
            A dataframe containing different gene/complex annotations.

        Notes
        -----
        There might be also a few miRNAs annotated. A vast majority of protein complex annotations are inferred
        from the annotations of the members: if all members carry the same annotation the complex inherits.
        """
        if proteins is None and resources is None and not force_full_download:
            raise ValueError(
                "Please specify `force_full_download=True` in order to download the full dataset."
            )
        res_info = (
            "all resources"
            if resources is None
            else f"the following resources: `{[resources] if isinstance(resources, str) else sorted(set(resources))}`"
        )
        inst = cls()

        if proteins is not None:
            if isinstance(proteins, str):
                proteins = (proteins,)
            proteins = sorted(set(proteins))

            logging.info(
                f"Downloading annotations for `{len(proteins)}` in `{_MAX_N_PROTS}` chunks from {res_info}"
            )

            return pd.concat(
                [
                    inst._get(
                        proteins=proteins[i * _MAX_N_PROTS : (i + 1) * _MAX_N_PROTS],
                        resources=resources,
                        **kwargs,
                    )
                    for i in range((len(proteins) // _MAX_N_PROTS) + 1)
                    if len(proteins[i * _MAX_N_PROTS : (i + 1) * _MAX_N_PROTS])
                ]
            )

        logging.info(f"Downloading annotations for all proteins from {res_info}")

        return inst._get(proteins=None, resources=resources, **kwargs)

    def _resource_filter(self, data: Mapping[str, Any], **_) -> bool:
        return True

    def _post_process(self, df: pd.DataFrame, **kwargs) -> pd.DataFrame:
        return df


__all__ = [Annotations]



from omnipath._core.requests._request import Enzsub, SignedPTMs
from omnipath._core.requests._complexes import Complexes
from omnipath._core.requests._intercell import Intercell
from omnipath._core.requests._annotations import Annotations



from typing import Union, Optional, Sequence

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


Strseq_t = Optional[Union[str, Sequence[str]]]
Organism_t = Literal["human", "mouse", "rat"]
License_t = Literal["academic", "commercial"]
Bool_t = Optional[bool]
Str_t = Optional[str]
Int_t = Optional[int]
None_t = type(None)



from io import BytesIO
from copy import copy
from typing import Any, Mapping, Callable, Optional
from hashlib import md5
from urllib.parse import urljoin
import json
import logging

from requests import Request, Session, PreparedRequest
from tqdm.auto import tqdm
from urllib3.util import Retry
from requests.adapters import HTTPAdapter

from omnipath._core.utils._options import Options
from omnipath.constants._pkg_constants import (
    UNKNOWN_SERVER_VERSION,
    Key,
    Format,
    Endpoint,
)


class Downloader:
    """
    Class which performs a GET request to the server in order to retrieve some remote resources.

    Also implements other behavior, such as retrying after some status codes.

    Parameters
    ----------
    opts
        Options. If `None`, :attr:`omnipath.options` are used.
    """

    def __init__(self, opts: Optional[Options] = None):
        if opts is None:
            from omnipath import options as opts

        if not isinstance(opts, Options):
            raise TypeError(
                f"Expected `opts` to be of type `Options`, found {type(opts).__name__}."
            )

        self._session = Session()
        self._options = copy(opts)  # this does not copy MemoryCache

        if self._options.num_retries > 0:
            adapter = HTTPAdapter(
                max_retries=Retry(
                    total=self._options.num_retries,
                    redirect=5,
                    allowed_methods=["HEAD", "GET", "OPTIONS"],
                    status_forcelist=[413, 429, 500, 502, 503, 504],
                    backoff_factor=1,
                )
            )
            self._session.mount("http://", adapter)
            self._session.mount("https://", adapter)

        logging.debug(f"Initialized `{self}`")

    @property
    def resources(self) -> Mapping[str, Mapping[str, Any]]:
        """Return the resources."""
        logging.debug("Fetching resources")
        return self.maybe_download(
            Endpoint.RESOURCES.s,
            params={Key.FORMAT.s: Format.JSON.s},
            callback=json.load,
            is_final=False,
        )

    def maybe_download(
        self,
        url: str,
        callback: Callable[[BytesIO], Any],
        params: Optional[Mapping[str, str]] = None,
        cache: bool = True,
        is_final: bool = True,
        **_,
    ) -> Any:
        """
        Fetch the data from the cache, if present, or download them from the ``url``.

        The key, under which is the download result saved, is the MD5 hash of the ``url``, including the ``params``.

        Parameters
        ----------
        url
            URL that is used to access the remote resources if the data is not found in the cache.
        callback
            Function applied on the downloaded data. Usually, this will return either a :class:`pandas.DataFrame`
            or a :class:`dict`.
        params
            Parameters of the `GET` request.
        cache
            Whether to save the files to the cache or not.
        is_final
            Whether ``url`` is final or should be prefixed with :paramref:`_options.url`.

        Returns
        -------
        :class:`typing.Any`
            The result of applying ``callback`` on the maybe downloaded data.
        """
        if not callable(callback):
            raise TypeError(
                f"Expected `callback` to be `callable`, found `{type(callback).__name__}`."
            )

        if not is_final:
            url = urljoin(self._options.url, url)

        req = self._session.prepare_request(
            Request(
                "GET",
                url,
                params=params,
                headers={"User-agent": "omnipathdb-user"},
            )
        )
        key = md5(bytes(req.url, encoding="utf-8")).hexdigest()

        if key in self._options.cache:
            logging.debug(f"Found data in cache `{self._options.cache}[{key!r}]`")
            res = self._options.cache[key]
        else:
            res = callback(self._download(req))
            if cache:
                logging.debug(f"Caching result to `{self._options.cache}[{key!r}]`")
                self._options.cache[key] = res
            else:
                logging.debug("Not caching the results")

        return res

    def _download(self, req: PreparedRequest) -> BytesIO:
        """
        Request the remote resources.

        Parameters
        ----------
        req
            `GET` request to perform.

        Returns
        -------
        :class:`io.BytesIO`
            File-like object containing the data. Usually a json- or csv-like data is present inside.
        """
        logging.info(f"Downloading data from `{req.url}`")

        handle = BytesIO()
        with self._session.send(
            req, stream=True, timeout=self._options.timeout
        ) as resp:
            resp.raise_for_status()
            total = resp.headers.get("content-length", None)

            with tqdm(
                unit="B",
                unit_scale=True,
                miniters=1,
                unit_divisor=1024,
                total=total if total is None else int(total),
                disable=not self._options.progress_bar,
            ) as t:
                for chunk in resp.iter_content(chunk_size=self._options.chunk_size):
                    t.update(len(chunk))
                    handle.write(chunk)

                handle.flush()
                handle.seek(0)

        return handle

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}[options={self._options}]>"

    def __repr__(self) -> str:
        return str(self)


def _get_server_version(options: Options) -> str:
    """Try and get the server version."""
    import re

    def callback(fp: BytesIO) -> str:
        """Parse the version."""
        return re.findall(
            r"\d+\.\d+.\d+", fp.getvalue().decode("utf-8"), flags=re.IGNORECASE
        )[0]

    try:
        if not options.autoload:
            raise ValueError("Autoload is disabled.")

        with Options.from_options(
            options,
            num_retries=0,
            timeout=0.1,
            cache=None,
            progress_bar=False,
            chunk_size=1024,
        ) as opt:
            return Downloader(opt).maybe_download(
                Endpoint.ABOUT.s,
                callback,
                params={Key.FORMAT.s: Format.TEXT.s},
                cache=False,
                is_final=False,
            )
    except Exception as e:
        logging.debug(f"Unable to get server version. Reason: `{e}`")

        return UNKNOWN_SERVER_VERSION



from io import StringIO
from urllib.parse import urljoin, quote_plus
import json

import pytest

import numpy as np
import pandas as pd

from omnipath import options
from omnipath.constants import Organism, InteractionDataset
from omnipath._core.requests import Intercell
from omnipath.constants._pkg_constants import Key, Endpoint
from omnipath._core.requests.interactions._utils import import_intercell_network
from omnipath._core.requests.interactions._interactions import (
    TFmiRNA,
    Dorothea,
    OmniPath,
    TFtarget,
    KinaseExtra,
    LigRecExtra,
    PathwayExtra,
    AllInteractions,
    Transcriptional,
    PostTranslational,
    miRNA,
    lncRNAmRNA,
)


class TestInteractions:
    def test_all_excluded_excluded(self):
        with pytest.raises(
            ValueError, match=r"After excluding `\d+` datasets, none were left."
        ):
            AllInteractions.get(exclude=list(InteractionDataset))

    def test_invalid_excluded_datasets(self):
        with pytest.raises(
            ValueError, match=r"Invalid value `foo` for `InteractionDataset`."
        ):
            AllInteractions.get(exclude="foo")

    def test_graph_empty(self):
        with pytest.raises(ValueError, match=r"No data were retrieved. Please"):
            AllInteractions.graph(pd.DataFrame())

    def test_graph_source_target(self):
        interaction = pd.DataFrame(
            {
                "source": ["alpha", "beta", "gamma"],
                "target": [0, 1, 0],
                "source_genesymbol": "bar",
                "target_genesymbol": "baz",
            }
        )
        src, tgt = AllInteractions._get_source_target_cols(interaction)

        assert src == "source_genesymbol"
        assert tgt == "target_genesymbol"

        src, tgt = AllInteractions._get_source_target_cols(
            interaction[
                interaction.columns.difference(
                    ["source_genesymbol", "target_genesymbol"]
                )
            ]
        )

        assert src == "source"
        assert tgt == "target"

    @pytest.mark.parametrize(
        "interaction",
        [
            PathwayExtra,
            KinaseExtra,
            LigRecExtra,
            miRNA,
            TFmiRNA,
            lncRNAmRNA,
            Dorothea,
            TFtarget,
            OmniPath,
            PostTranslational,
        ],
    )
    def test_resources(
        self, cache_backup, interaction, interaction_resources: bytes, requests_mock
    ):
        url = urljoin(options.url, Endpoint.RESOURCES.s)
        data = json.loads(interaction_resources)
        requests_mock.register_uri(
            "GET", f"{url}?format=json", content=interaction_resources
        )

        resources = interaction.resources()
        for resource in resources:
            assert {
                InteractionDataset(d)
                for d in data[resource][Key.QUERIES.s][
                    interaction._query_type.endpoint
                ][Key.DATASETS.s]
            } & interaction()._datasets
        assert requests_mock.called_once

    def test_invalid_organism(self):
        with pytest.raises(
            ValueError, match=r"Invalid value `foo` for `Organism`. Valid options are:"
        ):
            AllInteractions.get(**{Key.ORGANISM.s: "foo"})

    @pytest.mark.parametrize(
        "interaction",
        [
            PathwayExtra,
            KinaseExtra,
            LigRecExtra,
            miRNA,
            TFmiRNA,
            lncRNAmRNA,
            Dorothea,
            TFtarget,
            OmniPath,
            PostTranslational,
            AllInteractions,
        ],
    )
    def test_intercation_get(
        self, cache_backup, interaction, interaction_resources: bytes, requests_mock
    ):
        url = urljoin(options.url, interaction._query_type.endpoint)
        datasets = quote_plus(
            ",".join(sorted(d.value for d in interaction()._datasets))
        )
        fields = "fields=curation_effort%2Creferences%2Csources"
        if interaction is AllInteractions:
            fields += "%2Ctype"

        requests_mock.register_uri(
            "GET",
            f"{url}?datasets={datasets}&{fields}&format=tsv",
            content=interaction_resources,
        )

        _ = interaction.get()

        assert requests_mock.called_once

    @pytest.mark.parametrize("organisms", list(Organism))
    def test_valid_organism(
        self, cache_backup, organisms, requests_mock, interaction_resources
    ):
        url = urljoin(options.url, AllInteractions._query_type.endpoint)
        datasets = quote_plus(",".join(sorted(d.value for d in InteractionDataset)))
        requests_mock.register_uri(
            "GET",
            f"{url}?datasets={datasets}&fields=curation_effort%2Creferences%2Csources%2Ctype&"
            f"format=tsv&organisms={organisms.code}",
            content=interaction_resources,
        )

        _ = AllInteractions.get(organism=organisms, format="tsv")
        _ = AllInteractions.get(organisms=organisms.value, format="tsv")
        assert requests_mock.called_once

    def test_dorothea_params(self):
        params = Dorothea.params()

        assert "dorothea_levels" in params
        assert "dorothea_methods" in params
        assert "tfregulons_levels" not in params
        assert "tfregulons_methods" not in params
        assert Key.DATASETS.s not in params

    def test_tftarget_params(self):
        params = TFtarget.params()

        assert "dorothea_levels" not in params
        assert "dorothea_methods" not in params
        assert "tfregulons_levels" in params
        assert "tfregulons_methods" in params
        assert Key.DATASETS.s not in params

    @pytest.mark.parametrize(
        "interaction", [OmniPath, Transcriptional, AllInteractions]
    )
    def test_transcriptional_params(self, interaction):
        params = interaction.params()

        assert "dorothea_levels" in params
        assert "dorothea_methods" in params
        assert "tfregulons_levels" in params
        assert "tfregulons_methods" in params
        assert Key.DATASETS.s not in params

    @pytest.mark.parametrize(
        "interaction",
        [PathwayExtra, KinaseExtra, LigRecExtra, miRNA, TFmiRNA, lncRNAmRNA],
    )
    def test_rest_params(self, interaction):
        params = interaction.params()

        assert "dorothea_levels" not in params
        assert "dorothea_methods" not in params
        assert "tfregulons_levels" not in params
        assert "tfregulons_methods" not in params
        assert Key.DATASETS.s not in params


class TestUtils:
    def test_import_intercell_network(
        self,
        cache_backup,
        requests_mock,
        interactions_data: bytes,
        transmitters_data: bytes,
        receivers_data: bytes,
        import_intercell_result: pd.DataFrame,
    ):
        interactions_url = urljoin(options.url, AllInteractions._query_type.endpoint)
        intercell_url = urljoin(options.url, Intercell._query_type.endpoint)

        # interactions
        requests_mock.register_uri(
            "GET",
            f"{interactions_url}?datasets=omnipath&dorothea_levels=A&fields=curation_effort%2C"
            f"references%2Csources%2Ctype&format=tsv",
            content=interactions_data,
        )
        # transmitter
        requests_mock.register_uri(
            "GET",
            f"{intercell_url}?categories=ligand&causality=trans&format=tsv&scope=generic",
            content=transmitters_data,
        )
        # receiver
        requests_mock.register_uri(
            "GET",
            f"{intercell_url}?categories=receptor&causality=rec&format=tsv&scope=generic",
            content=receivers_data,
        )

        res = import_intercell_network(
            include=InteractionDataset.OMNIPATH,
            transmitter_params={"categories": "ligand"},
            interactions_params={"dorothea_levels": "A"},
            receiver_params={"categories": "receptor"},
        )

        assert isinstance(res, pd.DataFrame)
        np.testing.assert_array_equal(res.shape, import_intercell_result.shape)

        np.testing.assert_array_equal(res.index, import_intercell_result.index)
        np.testing.assert_array_equal(res.columns, import_intercell_result.columns)
        np.testing.assert_array_equal(res.dtypes, import_intercell_result.dtypes)
        np.testing.assert_array_equal(
            pd.isnull(res), pd.isnull(import_intercell_result)
        )
        np.testing.assert_array_equal(
            res.values[~pd.isnull(res)],
            import_intercell_result.values[~pd.isnull(import_intercell_result)],
        )
        assert len(requests_mock.request_history) == 3

    @pytest.mark.parametrize("which", ["interactions", "receivers", "transmitters"])
    def test_intercell_empty(
        self,
        which: str,
        cache_backup,
        requests_mock,
        interactions_data: bytes,
        transmitters_data: bytes,
        receivers_data: bytes,
    ):
        interactions_url = urljoin(options.url, AllInteractions._query_type.endpoint)
        intercell_url = urljoin(options.url, Intercell._query_type.endpoint)

        handle = StringIO()
        pd.DataFrame({"is_directed": []}).to_csv(handle, sep="\t", index=False)
        empty_data = bytes(handle.getvalue(), encoding="utf-8")

        if which == "interactions":
            interactions_data = empty_data
        elif which == "receivers":
            receivers_data = empty_data
        elif which == "transmitters":
            transmitters_data = empty_data
        else:
            raise AssertionError(which)

        # interactions
        requests_mock.register_uri(
            "GET",
            f"{interactions_url}?datasets=omnipath,pathwayextra&fields=curation_effort%2C"
            f"references%2Csources%2Ctype&format=tsv&resources=CellPhoneDB",
            content=interactions_data,
        )
        # transmitter
        requests_mock.register_uri(
            "GET",
            f"{intercell_url}?categories=ligand&causality=trans&format=tsv&scope=generic",
            content=transmitters_data,
        )
        # receiver
        requests_mock.register_uri(
            "GET",
            f"{intercell_url}?categories=receptor&causality=rec&format=tsv&scope=generic",
            content=receivers_data,
        )

        with pytest.raises(ValueError, match=rf"No {which} were retrieved. Please"):
            import_intercell_network(
                include=(InteractionDataset.OMNIPATH, InteractionDataset.PATHWAY_EXTRA),
                transmitter_params={"categories": "ligand"},
                interactions_params={"resources": "CellPhoneDB"},
                receiver_params={"categories": "receptor"},
            )



from omnipath._core.cache import clear_cache
from omnipath._core.utils import options  # from_first in isort is important here
from omnipath._core.downloader._downloader import _get_server_version
import omnipath.requests as requests
import omnipath.constants as constants
import omnipath.interactions as interactions

__author__ = ", ".join(["Michal Klein", "Dénes Türei"])
__maintainer__ = ", ".join(["Michal Klein", "Dénes Türei"])
__version__ = "1.0.4"
__email__ = "turei.denes@gmail.com"

try:
    from importlib_metadata import version  # Python < 3.8
except ImportError:
    from importlib.metadata import version  # Python = 3.8

from packaging.version import parse

__full_version__ = parse(version(__name__))
__full_version__ = (
    f"{__version__}+{__full_version__.local}" if __full_version__.local else __version__
)
__server_version__ = _get_server_version(options)

del parse, version, _get_server_version



from types import MethodType
from typing import *  # noqa: F401 F403 (because of the argspec factory)
from typing import Any, Dict, Union, Callable, Iterable, Optional
from inspect import Parameter, unwrap, isabstract
import inspect

import wrapt
import typing_extensions  # noqa: F401

import pandas as pd

from omnipath._core.utils._docs import d


@d.get_full_description(base="get")
@d.get_sections(base="get", sections=["Parameters", "Returns"])
def _get_helper(cls: type, **kwargs) -> pd.DataFrame:
    """
    Perform a request to the [OmniPath]_ web service.

    Parameters
    ----------
    kwargs
        Additional query parameters.

    Returns
    -------
    :class:`pandas.DataFrame`
        The result which depends the type of the request and the supplied parameters.
    """
    return cls()._get(**kwargs)


def _inject_api_method(
    clazz: type,
) -> None:
    """
    Create a decorator which does nothing except for modifying the function signature in the docstring.

    The function to be decorated must be a class method and is allowed only to have positional arguments,
    and variable keyword arguments (**kwargs).

    The resulting decorated function will containing only the positional arguments (including original type annotations)
    and possibly keyword only arguments. In this example signature might def fn(foo, bar, *, baz, quux),
    `baz` and `quux` are the keyword only arguments.

    Parameters
    ----------
    clazz
        The class for which to create the query. Must not be abstract.

    Returns
    -------
    :class:`callable`
        The decorator as described above.
    """

    def argspec_factory(orig_fn: Callable) -> Callable:
        orig_params = inspect.signature(orig_fn).parameters
        # maintain the original signature if the subclass has overriden the method
        # this will lose the docstring of the original function
        parameters = {
            k: v
            for k, v in orig_params.items()
            if k != "cls"
            and v.kind in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD)
        }
        annotations = {
            k: v for k, v in clazz._annotations().items() if k not in parameters
        }

        sig = inspect.signature(lambda _: _)
        sig = sig.replace(
            parameters=[Parameter("cls", kind=Parameter.POSITIONAL_ONLY)]
            + list(parameters.values())
            + [
                Parameter(k, kind=Parameter.KEYWORD_ONLY, annotation=a)
                for k, a in sorted(annotations.items())
            ]
            + [Parameter("kwargs", kind=Parameter.VAR_KEYWORD)]
        )
        # modify locals() for argspec factory
        import omnipath  # noqa: F401

        NoneType, pandas = type(None), pd

        exec(
            f"def adapter{sig}: pass".replace(" /,", ""),
            globals(),
            locals(),
        )
        return locals()["adapter"]

    if not isinstance(clazz, type):
        raise TypeError(
            f"Expected `clazz` to be a type, found `{type(clazz).__name__}`."
        )

    if isabstract(clazz):
        return

    @wrapt.decorator(adapter=wrapt.adapter_factory(argspec_factory))
    def wrapper(wrapped, _instance, args, kwargs):
        return wrapped(*args, **kwargs)

    if hasattr(clazz, "get") and not hasattr(clazz.get, "__wrapped__"):
        # overriding in subclass
        clazz.get = wrapper(unwrap(clazz.get))
        return

    clazz.get = wrapper(MethodType(_get_helper, clazz))


def _inject_params(
    params: Dict[str, Any], key: str, value: Optional[Union[str, Iterable[str]]]
) -> None:
    if value is None:
        return
    value = {value} if isinstance(value, str) else set(value)

    old_value = params.pop(key, None)
    if old_value is None:
        params[key] = value
        return

    old_value = {old_value} if isinstance(old_value, str) else set(old_value)

    params[key] = value | old_value


def _split_unique_join(data: pd.Series, func: Optional[Callable] = None) -> pd.Series:
    mask = ~pd.isnull(data.astype("string"))
    data = data[mask]
    data = data.str.split(";")

    if func is None:
        data = data.apply(
            lambda row: ";".join(sorted(set(map(str, row))))
            if isinstance(row, Iterable)
            else row
        )
    else:
        data = data.apply(func)

    res = pd.Series([None] * len(mask))
    res.loc[mask] = data

    return res


def _strip_resource_label(
    data: pd.Series, func: Optional[Callable] = None
) -> pd.Series:
    return _split_unique_join(
        _split_unique_join(data.str.replace(r"[-\w]*:?(\d+)", r"\1")), func=func
    )


_ERROR_EMPTY_FMT = (
    "No {obj} were retrieved. Please check if supplying valid parameter values."
)



from io import BytesIO, StringIO
from urllib.parse import urljoin
import logging

import pytest

import numpy as np
import pandas as pd

from omnipath import options as opt
from omnipath._core.utils._options import Options
from omnipath.constants._pkg_constants import UNKNOWN_SERVER_VERSION, Endpoint
from omnipath._core.downloader._downloader import Downloader, _get_server_version


class TestDownloader:
    def test_options_wrong_type(self):
        with pytest.raises(TypeError):
            Downloader("foobar")

    def test_str_repr(self, options: Options):
        d = Downloader(options)

        assert str(d) == f"<{d.__class__.__name__}[options={options}]>"
        assert repr(d) == f"<{d.__class__.__name__}[options={options}]>"

    def test_initialize_local_options(self, options: Options):
        options.password = "foo"
        options.timeout = 1337
        d = Downloader(options)

        assert d._options is not options
        assert str(d._options) == str(options)
        assert str(d._options) != str(opt)

        options.password = "bar"
        assert d._options.password == "foo"

    def test_initialize_global_options(self):
        d = Downloader()

        assert d._options is not opt
        assert str(d._options) == str(opt)

    def test_resources_cached_values(self, downloader: Downloader, requests_mock):
        data = {"foo": "bar", "42": 1337}
        requests_mock.register_uri(
            "GET", urljoin(downloader._options.url, Endpoint.RESOURCES.s), json=data
        )

        assert downloader.resources == data
        assert requests_mock.called_once

        assert downloader.resources == data
        assert requests_mock.called_once

    def test_resources_no_cached_values(self, downloader: Downloader, requests_mock):
        data = {"foo": "bar", "42": 1337}
        requests_mock.register_uri(
            "GET", urljoin(downloader._options.url, Endpoint.RESOURCES.s), json=data
        )

        assert downloader.resources == data
        assert requests_mock.called_once

        downloader._options.cache.clear()

        assert downloader.resources == data
        assert len(requests_mock.request_history) == 2

    def test_maybe_download_not_callable(self, downloader: Downloader):
        with pytest.raises(TypeError):
            downloader.maybe_download("foo", callback=None)

    def test_maybe_download_wrong_callable(
        self, downloader: Downloader, requests_mock, csv_data: bytes
    ):
        url = urljoin(downloader._options.url, "foobar")
        requests_mock.register_uri("GET", url, content=csv_data)

        with pytest.raises(ValueError, match=r"Expected object or value"):
            downloader.maybe_download(url, callback=pd.read_json)

    def test_maybe_download_passes_params(
        self, downloader: Downloader, requests_mock, csv_data: bytes
    ):
        csv_url = urljoin(downloader._options.url, "foobar/?format=csv")
        csv_df = pd.read_csv(BytesIO(csv_data))
        json_url = urljoin(downloader._options.url, "foobar/?format=json")
        json_handle = StringIO()
        csv_df.to_json(json_handle)

        requests_mock.register_uri("GET", csv_url, content=csv_data)
        requests_mock.register_uri(
            "GET", json_url, content=bytes(json_handle.getvalue(), encoding="utf-8")
        )

        res1 = downloader.maybe_download(csv_url, callback=pd.read_csv)
        res2 = downloader.maybe_download(csv_url, callback=pd.read_csv)

        assert res1 is res2
        assert requests_mock.called_once
        np.testing.assert_array_equal(res1.index, csv_df.index)
        np.testing.assert_array_equal(res1.columns, csv_df.columns)
        np.testing.assert_array_equal(res1.values, csv_df.values)

        res1 = downloader.maybe_download(json_url, callback=pd.read_json)
        res2 = downloader.maybe_download(json_url, callback=pd.read_json)

        assert res1 is res2
        assert len(requests_mock.request_history) == 2
        np.testing.assert_array_equal(res1.index, csv_df.index)
        np.testing.assert_array_equal(res1.columns, csv_df.columns)
        np.testing.assert_array_equal(res1.values, csv_df.values)

    def test_maybe_download_no_cache(
        self, downloader: Downloader, requests_mock, csv_data: bytes
    ):
        url = urljoin(downloader._options.url, "foobar")
        requests_mock.register_uri("GET", url, content=csv_data)

        res1 = downloader.maybe_download(url, callback=pd.read_csv)
        downloader._options.cache.clear()
        res2 = downloader.maybe_download(url, callback=pd.read_csv)

        assert res1 is not res2
        assert len(requests_mock.request_history) == 2
        np.testing.assert_array_equal(res1.index, res2.index)
        np.testing.assert_array_equal(res1.columns, res2.columns)
        np.testing.assert_array_equal(res1.values, res2.values)

    def test_maybe_download_is_not_final(
        self, downloader: Downloader, requests_mock, csv_data: bytes
    ):
        endpoint = "barbaz"
        url = urljoin(downloader._options.url, endpoint)
        requests_mock.register_uri("GET", url, content=csv_data)
        csv_df = pd.read_csv(BytesIO(csv_data))

        res = downloader.maybe_download(endpoint, callback=pd.read_csv, is_final=False)

        assert requests_mock.called_once
        np.testing.assert_array_equal(res.index, csv_df.index)
        np.testing.assert_array_equal(res.columns, csv_df.columns)
        np.testing.assert_array_equal(res.values, csv_df.values)

    def test_get_server_version_not_decodable(
        self, options: Options, requests_mock, caplog
    ):
        url = urljoin(options.url, Endpoint.ABOUT.s)
        options.autoload = True
        requests_mock.register_uri(
            "GET", f"{url}?format=text", content=bytes("foobarbaz", encoding="utf-8")
        )

        with caplog.at_level(logging.DEBUG):
            version = _get_server_version(options)

        assert requests_mock.called_once
        assert (
            "Unable to get server version. Reason: `list index out of range`"
            in caplog.text
        )
        assert version == UNKNOWN_SERVER_VERSION

    def test_get_server_version_no_autoload(
        self, options: Options, requests_mock, caplog
    ):
        url = urljoin(options.url, Endpoint.ABOUT.s)
        options.autoload = False
        requests_mock.register_uri("GET", f"{url}?format=text", text="foobarbaz")

        with caplog.at_level(logging.DEBUG):
            version = _get_server_version(options)

        assert not requests_mock.called_once
        assert (
            "Unable to get server version. Reason: `Autoload is disabled.`"
            in caplog.text
        )
        assert version == UNKNOWN_SERVER_VERSION

    def test_get_server_version(self, options: Options, requests_mock):
        url = urljoin(options.url, Endpoint.ABOUT.s)
        options.autoload = True
        requests_mock.register_uri(
            "GET",
            f"{url}?format=text",
            content=bytes("foo bar baz\nversion: 42.1337.00", encoding="utf-8"),
        )

        version = _get_server_version(options)

        assert requests_mock.called_once
        assert version == "42.1337.00"



from io import StringIO
from abc import ABC
from copy import deepcopy
from shutil import copy
from typing import Optional
from inspect import isclass
from pathlib import Path
from collections import defaultdict
from urllib.parse import urljoin
import json
import pickle
import logging

import pytest
import requests

import numpy as np
import pandas as pd

from omnipath.constants import InteractionDataset
from omnipath._core.cache._cache import MemoryCache
from omnipath._core.query._query import QueryType
from omnipath._core.utils._options import Options
from omnipath.constants._pkg_constants import DEFAULT_OPTIONS, Key, Endpoint
from omnipath._core.downloader._downloader import Downloader
import omnipath as op


def pytest_addoption(parser):
    parser.addoption(
        "--test-server",
        dest="test_server",
        action="store_true",
        help="Whether to also test the server connection.",
    )


@pytest.fixture(scope="function")
def options() -> "Options":
    opt = Options.from_config()
    opt.cache = None
    opt.progress_bar = False
    return opt


@pytest.fixture(scope="function")
def config_backup(tmpdir):
    copy(Options.config_path, tmpdir / "config.ini")
    yield
    copy(tmpdir / "config.ini", Options.config_path)


@pytest.fixture(scope="function")
def cache_backup():
    import omnipath as op

    cache = deepcopy(op.options.cache)
    pb = op.options.progress_bar
    op.options.cache = MemoryCache()
    op.options.progress_bar = False
    yield
    op.options.cache = cache
    op.options.progress_bar = pb


@pytest.fixture(scope="class")
def server_url():
    import omnipath as op

    cache = deepcopy(op.options.cache)
    pb = op.options.progress_bar
    url = op.options.url
    cd = op.options.convert_dtypes

    op.options.cache = MemoryCache()
    op.options.progress_bar = False
    op.options.url = DEFAULT_OPTIONS.url
    op.options.convert_dtypes = True
    yield
    op.options.cache = cache
    op.options.progress_bar = pb
    op.options.url = url
    op.options.convert_dtypes = cd


@pytest.fixture(scope="function")
def downloader(options) -> "Downloader":
    return Downloader(options)


@pytest.fixture(scope="session")
def csv_data() -> bytes:
    str_handle = StringIO()
    pd.DataFrame({"foo": range(5), "bar": "baz", "quux": 42}).to_csv(str_handle)

    return bytes(str_handle.getvalue(), encoding="utf-8")


@pytest.fixture(scope="session")
def tsv_data() -> bytes:
    str_handle = StringIO()
    pd.DataFrame(
        {
            "foo": range(5),
            "components_genesymbols": "foo",
            "quux": 42,
            "modification": "bar",
        }
    ).to_csv(str_handle, sep="\t")

    return bytes(str_handle.getvalue(), encoding="utf-8")


@pytest.fixture(scope="session")
def intercell_data() -> bytes:
    data = {}
    data[Key.PARENT.s] = [42, 1337, 24, 42]
    data[Key.CATEGORY.s] = ["foo", "bar", "bar", "foo"]

    return bytes(json.dumps(data), encoding="utf-8")


@pytest.fixture(scope="session")
def resources() -> bytes:
    data = defaultdict(dict)
    data["foo"][Key.QUERIES.s] = {
        QueryType.INTERCELL.endpoint: {Key.GENERIC_CATEGORIES.s: ["42"]}
    }
    data["bar"][Key.QUERIES.s] = {
        QueryType.INTERCELL.endpoint: {Key.GENERIC_CATEGORIES.s: ["42", "13"]}
    }
    data["baz"][Key.QUERIES.s] = {
        QueryType.INTERCELL.endpoint: {Key.GENERIC_CATEGORIES.s: ["24"]}
    }
    data["quux"][Key.QUERIES.s] = {
        QueryType.ENZSUB.endpoint: {Key.GENERIC_CATEGORIES.s: ["24"]}
    }

    return bytes(json.dumps(data), encoding="utf-8")


@pytest.fixture(scope="session")
def interaction_resources() -> bytes:
    data = defaultdict(dict)
    for i, d in enumerate(InteractionDataset):
        data[f"d_{i}"][Key.QUERIES.s] = {
            QueryType.INTERACTIONS.endpoint: {Key.DATASETS.s: [d.value]}
        }

    return bytes(json.dumps(data), encoding="utf-8")


@pytest.fixture(scope="session")
def complexes() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "components_genesymbols": [
                "foo",
                "bar_baz_quux",
                "baz_bar",
                "bar_quux_foo",
            ],
            "dummy": 42,
        }
    )


@pytest.fixture(scope="session")
def interactions_data() -> bytes:
    str_handle = StringIO()
    with open(Path("tests") / "_data" / "interactions.pickle", "rb") as fin:
        data: pd.DataFrame = pickle.load(fin)

    data.to_csv(str_handle, sep="\t", index=False)

    return bytes(str_handle.getvalue(), encoding="utf-8")


@pytest.fixture(scope="session")
def transmitters_data() -> bytes:
    str_handle = StringIO()
    with open(Path("tests") / "_data" / "transmitters.pickle", "rb") as fin:
        data: pd.DataFrame = pickle.load(fin)

    data.to_csv(str_handle, sep="\t", index=False)

    return bytes(str_handle.getvalue(), encoding="utf-8")


@pytest.fixture(scope="session")
def receivers_data() -> bytes:
    str_handle = StringIO()
    with open(Path("tests") / "_data" / "receivers.pickle", "rb") as fin:
        data: pd.DataFrame = pickle.load(fin)

    data.to_csv(str_handle, sep="\t", index=False)

    return bytes(str_handle.getvalue(), encoding="utf-8")


@pytest.fixture(scope="session")
def import_intercell_result() -> pd.DataFrame:
    with open(Path("tests") / "_data" / "import_intercell_result.pickle", "rb") as fin:
        return pickle.load(fin)


@pytest.fixture(scope="session")
def string_series() -> pd.Series:
    return pd.Series(["foo:123", "bar:45;baz", None, "baz:67;bar:67", "foo;foo;foo"])


def _can_import_omnipathR() -> Optional["rpy2.robjects.packages.Package"]:  # noqa: F821
    try:
        from packaging import version
        from rpy2.robjects.packages import PackageNotInstalledError, importr
        import rpy2

        try:
            from importlib_metadata import version as get_version
        except ImportError:
            # >=Python3.8
            from importlib.metadata import version as get_version

        try:
            assert version.parse(get_version(rpy2.__name__)) >= version.parse("3.3.0")
            mod = importr("OmnipathR")
            logging.error("Succesfully loaded `OmnipathR`")
            return mod
        except (PackageNotInstalledError, AssertionError) as err:
            logging.error(f"Unable to import `OmnipathR`. Reason: `{err}`.")

    except ImportError as err:
        logging.error(f"Unable to import `rpy2`. Reason: `{err}`.")

    return None


@pytest.fixture(scope="session")
def omnipathr(request):
    url = urljoin(DEFAULT_OPTIONS.url, Endpoint.ABOUT.s)

    if not request.config.getoption("test_server", default=False, skip=True):
        logging.error("Testing using the server is disabled.")
        return None
    try:
        resp = requests.get(url)
        status_code = resp.status_code
    except Exception as e:
        logging.error(f"Unable to contact the server at `{url}`. Reason: `{e}`")
        return None

    if status_code != 200:
        logging.error(
            f"Unable to contact the server at `{url}`. Status code: `{status_code}`"
        )
        return None

    return _can_import_omnipathR()


@pytest.fixture(autouse=True, scope="class")
def _inject_omnipath(request, omnipathr, server_url):
    if isclass(request.cls) and issubclass(request.cls, RTester):
        if omnipathr is None:
            pytest.skip("Unable to import `OmnipathR`.")
        from rpy2.robjects import pandas2ri

        # at this point, we know rpy2 can be imported, thanks to the `omnipathr` fixture
        # do not change the activation order
        pandas2ri.activate()
        request.cls.omnipathr = omnipathr


class RTester(ABC):
    def test_resources(self):
        expected = sorted(self.omnipathr.get_intercell_resources())
        actual = sorted(op.requests.Intercell.resources())

        np.testing.assert_array_equal(expected, actual)



"""
Based on gist.github.com/MantasVaitkunas/7c16de233812adcb7028
"""

import warnings

__licence__ = "BSD (3 clause)"


def html_page_context(app, _pagename, templatename, context, doctree):
    # doctree is None - otherwise viewcode fails
    if templatename != "page.html" or doctree is None:
        return

    if not app.config.github_repo:
        warnings.warn("`github_repo` not specified")
        return

    # For sphinx_rtd_theme.
    context["display_github"] = True
    context["github_user"] = "saezlab"
    context["github_version"] = "master"
    context["github_repo"] = app.config.github_repo
    context["conf_py_path"] = "/docs/source/"


def setup(app):
    app.add_config_value("github_repo", "", True)
    app.connect("html-page-context", html_page_context)



from typing import Any, Dict, Mapping, Optional

import pandas as pd

from omnipath.constants._constants import InteractionDataset
from omnipath._core.requests._utils import _ERROR_EMPTY_FMT
from omnipath._core.requests._intercell import Intercell
from omnipath._core.requests.interactions._interactions import (
    Datasets_t,
    AllInteractions,
)


def _to_dict(mapping: Optional[Mapping[Any, Any]]) -> Dict[Any, Any]:
    return {} if mapping is None else dict(mapping)


def _swap_undirected(df: pd.DataFrame) -> pd.DataFrame:
    if "is_directed" not in df.columns:
        raise KeyError(f"Key `'is_directed'` not found in `{list(df.columns)}`.")

    directed = df.pop("is_directed")

    undirected = df.loc[~directed, :]
    if undirected.empty:
        return df

    undirected_swapped = undirected.copy()
    undirected_swapped[["source", "target"]] = undirected[["target", "source"]]

    if "source_genesymbol" in undirected:
        undirected_swapped[["source_genesymbol", "target_genesymbol"]] = undirected[
            ["target_genesymbol", "source_genesymbol"]
        ]
    if "ncbi_tax_id_source" in undirected.columns:
        undirected_swapped[["ncbi_tax_id_source", "ncbi_tax_id_target"]] = undirected[
            ["ncbi_tax_id_target", "ncbi_tax_id_source"]
        ]

    return pd.concat(
        [directed, undirected, undirected_swapped], axis=0, ignore_index=True
    )


def import_intercell_network(
    include: Datasets_t = (
        InteractionDataset.OMNIPATH,
        InteractionDataset.PATHWAY_EXTRA,
        InteractionDataset.KINASE_EXTRA,
        InteractionDataset.LIGREC_EXTRA,
    ),
    interactions_params: Optional[Mapping[str, Any]] = None,
    transmitter_params: Optional[Mapping[str, Any]] = None,
    receiver_params: Optional[Mapping[str, Any]] = None,
) -> pd.DataFrame:
    """
    Import intercellular network combining intercellular annotations and protein interactions.

    First, it imports a network of protein-protein interactions. Then, it retrieves annotations about the proteins
    intercellular communication roles, once for the transmitter (delivering information from the expressing cell) and
    second, the receiver (receiving signal and relaying it towards the expressing cell) side.

    These 3 queries can be customized by providing parameters which will be passed to
    :meth:`omnipath.interactions.OmniPath.get` for the network and :meth:`omnipath.requests.Intercell`
    for the annotations.

    Finally the 3 :class:`pandas.DataFrame` are combined in a way that the source proteins in each interaction annotated
    by the transmitter, and the target proteins by the receiver categories. If undirected interactions present
    (these are disabled by default) they will be duplicated, i.e. both partners can be both receiver and transmitter.

    Parameters
    ----------
    include
        Interaction datasets to include for :meth:`omnipath.interactions.AllInteractions.get`.
    interactions_params
        Parameters for the :meth:`omnipath.interactions.AllInteractions.get`.
    transmitter_params
        Parameters defining the transmitter side of intercellular connections.
        See :meth:`omnipath.interactions.AllInteractions.params` for available values.
    receiver_params
        Parameters defining the receiver side of intercellular connections.
        See :meth:`omnipath.interactions.AllInteractions.params` for available values.

    Returns
    -------
    :class:`pandas.DataFrame`
        A dataframe containing information about protein-protein interactions and the inter-cellular roles
        of the proteins involved in those interactions.
    """
    interactions_params = _to_dict(interactions_params)
    transmitter_params = _to_dict(transmitter_params)
    receiver_params = _to_dict(receiver_params)

    # TODO: this should be refactored as: QueryType.INTERCELL("scope").param, etc. (also in many other places)
    transmitter_params.setdefault("causality", "trans")
    transmitter_params.setdefault("scope", "generic")
    receiver_params.setdefault("causality", "rec")
    receiver_params.setdefault("scope", "generic")

    interactions = AllInteractions.get(include=include, **interactions_params)
    if interactions.empty:
        raise ValueError(_ERROR_EMPTY_FMT.format(obj="interactions"))
    interactions = _swap_undirected(interactions)

    transmitters = Intercell.get(**transmitter_params)
    if transmitters.empty:
        raise ValueError(_ERROR_EMPTY_FMT.format(obj="transmitters"))
    receivers = Intercell.get(**receiver_params)
    if receivers.empty:
        raise ValueError(_ERROR_EMPTY_FMT.format(obj="receivers"))

    # fmt: off
    intracell = ['intracellular_intercellular_related', 'intracellular']
    transmitters = transmitters.loc[~transmitters["parent"].isin(intracell), :]
    transmitters.rename(columns={"source": "category_source"}, inplace=True)
    # this makes it 3x as fast during groupby, since all of these are categories
    # it's mostly because groupby needs observed=True + using string object (numpy) vs "string"
    transmitters[["category", "parent", "database"]] = transmitters[["category", "parent", "database"]].astype(str)

    receivers = receivers.loc[~receivers["parent"].isin(intracell), :]
    receivers.rename(columns={"source": "category_source"}, inplace=True)
    receivers[["category", "parent", "database"]] = receivers[["category", "parent", "database"]].astype(str)

    res = pd.merge(interactions, transmitters, left_on="source", right_on="uniprot", how="inner")
    if res.empty:
        raise ValueError("No values are left after merging interactions and transmitters.")
    gb = res.groupby(["category", "parent", "source", "target"], as_index=False)
    # fmt: on

    res = gb.nth(0).copy()  # much faster than 1st
    res["database"] = gb["database"].apply(";".join)["database"].astype(str)

    res = pd.merge(
        res,
        receivers,
        how="inner",
        left_on="target",
        right_on="uniprot",
        suffixes=["_intercell_source", "_intercell_target"],
    )
    if res.empty:
        raise ValueError("No values are left after merging interactions and receivers.")
    gb = res.groupby(
        [
            "category_intercell_source",
            "parent_intercell_source",
            "source",
            "target",
            "category_intercell_target",
            "parent_intercell_target",
        ],
        as_index=False,
    )

    res = gb.nth(0).copy()
    res["database_intercell_target"] = (
        gb["database_intercell_target"]
        .apply(";".join)["database_intercell_target"]
        .astype(str)
    )

    # retype back as categories
    for col in ["category", "parent"]:
        for suffix in ["_intercell_source", "_intercell_target"]:
            res[f"{col}{suffix}"] = res[f"{col}{suffix}"].astype("category")

    return res.reset_index(drop=True)



from io import StringIO
from typing import Iterable, _GenericAlias
from urllib.parse import urljoin
import json
import logging

import pytest

from pandas.api.types import is_object_dtype, is_categorical_dtype
import numpy as np
import pandas as pd

from omnipath import options
from omnipath.requests import Enzsub, Complexes, Intercell, Annotations
from omnipath._core.requests import SignedPTMs
from omnipath._core.query._query import EnzsubQuery
from omnipath._core.requests._utils import _split_unique_join, _strip_resource_label
from omnipath.constants._pkg_constants import Key, Endpoint
from omnipath._core.requests._annotations import _MAX_N_PROTS


class TestEnzsub:
    def test_str_repr(self):
        assert str(Enzsub()) == f"<{Enzsub().__class__.__name__}>"
        assert repr(Enzsub()) == f"<{Enzsub().__class__.__name__}>"

    def test_params_no_org_genesymbol(self):
        params = Enzsub.params()

        assert Key.ORGANISM.value not in params
        assert Key.GENESYMBOLS.value in params

        for k, valid in params.items():
            if isinstance(valid, Iterable):
                np.testing.assert_array_equal(
                    list(set(valid)), list(set(EnzsubQuery(k).valid))
                )
            else:
                assert valid == EnzsubQuery(k).valid

    def test_resources(self, cache_backup, requests_mock, resources: bytes):
        url = urljoin(options.url, Endpoint.RESOURCES.s)
        requests_mock.register_uri("GET", f"{url}?format=json", content=resources)

        res = Enzsub.resources()

        assert res == ("quux",)
        assert requests_mock.called_once

    def test_invalid_params(self):
        with pytest.raises(ValueError, match=r"Invalid value `foo` for `EnzsubQuery`."):
            Enzsub.get(foo="bar")

    def test_invalid_license(self):
        with pytest.raises(ValueError, match=r"Invalid value `bar` for `License`."):
            Enzsub.get(license="bar")

    def test_invalid_format(self):
        with pytest.raises(ValueError, match=r"Invalid value `bar` for `Format`."):
            Enzsub.get(format="bar")

    def test_valid_params(self, cache_backup, requests_mock, tsv_data: bytes, caplog):
        url = urljoin(options.url, Enzsub._query_type.endpoint)
        df = pd.read_csv(StringIO(tsv_data.decode("utf-8")), sep="\t")
        requests_mock.register_uri(
            "GET",
            f"{url}?fields=curation_effort%2Creferences%2Csources&format=tsv&license=academic",
            content=tsv_data,
        )

        with caplog.at_level(logging.WARNING):
            res = Enzsub.get(license="academic", format="text")

        assert f"Invalid `{Key.FORMAT.s}='text'`" in caplog.text
        np.testing.assert_array_equal(res.index, df.index)
        np.testing.assert_array_equal(res.columns, df.columns)
        np.testing.assert_array_equal(res.values, df.values)
        assert requests_mock.called_once

    def test_annotations(self):
        assert set(Enzsub._annotations().keys()) == {e.param for e in EnzsubQuery}
        assert all(
            isinstance(a, (_GenericAlias, type)) for a in Enzsub._annotations().values()
        )

    def test_docs(self):
        assert set(Enzsub._docs().keys()) == {e.param for e in EnzsubQuery}
        assert all(d is None for d in Enzsub._docs().values())

    def test_invalid_organism(self, cache_backup, requests_mock, tsv_data: bytes):
        url = urljoin(options.url, Enzsub._query_type.endpoint)
        requests_mock.register_uri(
            "GET",
            f"{url}?fields=curation_effort%2Creferences%2Csources&format=tsv",
            content=tsv_data,
        )
        with pytest.raises(
            ValueError,
            match=r"Invalid value `foobarbaz` for `Organism`. Valid options are:",
        ):
            Enzsub.get(organism="foobarbaz")

        assert not requests_mock.called_once

    def test_genesymbols_dont_matter(
        self, cache_backup, requests_mock, tsv_data: bytes
    ):
        url = urljoin(options.url, Enzsub._query_type.endpoint)
        requests_mock.register_uri(
            "GET",
            f"{url}?fields=curation_effort%2Creferences%2Csources&format=tsv",
            content=tsv_data,
        )
        _ = Enzsub.get(genesymbol=True)

        assert requests_mock.called_once

    def test_field_injection(self, cache_backup, requests_mock, tsv_data: bytes):
        url = urljoin(options.url, Enzsub._query_type.endpoint)
        requests_mock.register_uri(
            "GET",
            f"{url}?fields=Alpha%2Cbeta%2Ccuration_effort%2Creferences%2Csources&format=tsv",
            content=tsv_data,
        )
        _ = Enzsub.get(fields=("beta", "Alpha", "Alpha"))

        assert requests_mock.called_once

    def test_no_dtype_conversion(self, cache_backup, requests_mock, tsv_data: bytes):
        url = urljoin(options.url, Enzsub._query_type.endpoint)
        options.convert_dtypes = False

        requests_mock.register_uri(
            "GET",
            f"{url}?fields=curation_effort%2Creferences%2Csources&format=tsv",
            content=tsv_data,
        )

        res = Enzsub.get()
        assert is_object_dtype(res["modification"])

        options.convert_dtypes = True

        res = Enzsub.get()
        assert is_categorical_dtype(res["modification"])
        assert requests_mock.called_once


class TestIntercell:
    def test_resources_wrong_type(self):
        with pytest.raises(TypeError):
            Intercell.resources(42)

    def test_resources_no_generic_resources(self):
        with pytest.raises(
            ValueError, match=r"No generic categories have been selected."
        ):
            Intercell.resources([])

    def test_resources_no_generic(self, cache_backup, requests_mock, resources: bytes):
        url = urljoin(options.url, Endpoint.RESOURCES.s)
        requests_mock.register_uri("GET", f"{url}?format=json", content=resources)

        res = Intercell.resources()

        assert res == ("bar", "baz", "foo")
        assert requests_mock.called_once

    def test_resources_generic(self, cache_backup, requests_mock, resources: bytes):
        url = urljoin(options.url, Endpoint.RESOURCES.s)
        requests_mock.register_uri("GET", f"{url}?format=json", content=resources)

        res = Intercell.resources(generic_categories=["42"])
        assert res == ("bar", "foo")

        res = Intercell.resources(generic_categories="24")
        assert res == ("baz",)

        res = Intercell.resources(generic_categories="foobarbaz")
        assert res == ()
        assert requests_mock.called_once  # caching

    def test_categories(self, cache_backup, requests_mock, intercell_data: bytes):
        url = urljoin(options.url, Key.INTERCELL_SUMMARY.s)
        data = json.loads(intercell_data)
        requests_mock.register_uri("GET", f"{url}?format=json", content=intercell_data)

        res = Intercell.categories()

        assert res == tuple(sorted(set(map(str, data[Key.CATEGORY.s]))))
        assert requests_mock.called_once

    def test_generic_categories(
        self, cache_backup, requests_mock, intercell_data: bytes
    ):
        url = urljoin(options.url, Key.INTERCELL_SUMMARY.s)
        data = json.loads(intercell_data)
        requests_mock.register_uri("GET", f"{url}?format=json", content=intercell_data)

        res = Intercell.generic_categories()

        assert res == tuple(sorted(set(map(str, data[Key.PARENT.s]))))
        assert requests_mock.called_once

    def test_password_from_options(
        self, cache_backup, requests_mock, intercell_data: bytes
    ):
        old_pwd = options.password
        options.password = "foobar"

        url = urljoin(options.url, Intercell._query_type.endpoint)
        requests_mock.register_uri(
            "GET",
            f"{url}?format=tsv&password=foobar",
            content=intercell_data,
        )

        _ = Intercell.get()
        options.password = old_pwd

        assert requests_mock.called_once

    def test_password_from_function_call(
        self, cache_backup, requests_mock, intercell_data: bytes
    ):
        old_pwd = options.password
        options.password = "foobar"

        url = urljoin(options.url, Intercell._query_type.endpoint)
        requests_mock.register_uri(
            "GET",
            f"{url}?format=tsv&password=bazquux",
            content=intercell_data,
        )

        _ = Intercell.get(password="bazquux")
        options.password = old_pwd

        assert requests_mock.called_once


class TestComplex:
    def test_complex_genes_wrong_dtype(self):
        with pytest.raises(TypeError):
            Complexes.complex_genes("foo", complexes=42)

    def test_comples_genes_empty_complexes(self, caplog):
        df = pd.DataFrame()
        with caplog.at_level(logging.WARNING):
            res = Complexes.complex_genes("foo", complexes=df)

        assert res is df
        assert "Complexes are empty" in caplog.text

    def test_complex_genes_no_column(self):
        with pytest.raises(KeyError):
            Complexes.complex_genes("foo", complexes=pd.DataFrame({"foo": range(10)}))

    def test_complex_genes_no_genes(self):
        with pytest.raises(ValueError, match=r"No genes have been selected."):
            Complexes.complex_genes([], complexes=None)

    def test_complex_genes_complexes_not_specified(
        self, cache_backup, requests_mock, tsv_data: bytes
    ):
        url = urljoin(options.url, Complexes._query_type.endpoint)
        df = pd.read_csv(StringIO(tsv_data.decode("utf-8")), sep="\t")
        requests_mock.register_uri("GET", f"{url}?format=tsv", content=tsv_data)

        res = Complexes.complex_genes("fooo")

        np.testing.assert_array_equal(res.columns, df.columns)
        assert res.empty

    def test_complexes_complexes_specified(self, complexes: pd.DataFrame):
        res = Complexes.complex_genes("foo", complexes=complexes, total_match=False)

        assert isinstance(res, pd.DataFrame)
        assert res.shape == (2, 2)
        assert set(res.columns) == {"components_genesymbols", "dummy"}
        assert all(
            any(v in "foo" for v in vs.split("_"))
            for vs in res["components_genesymbols"]
        )

    def test_complexes_total_match(self, complexes: pd.DataFrame):
        res = Complexes.complex_genes(
            ["bar", "baz"], complexes=complexes, total_match=True
        )

        assert res.shape == (1, 2)
        assert all(
            all(v in ("bar", "baz") for v in vs.split("_"))
            for vs in res["components_genesymbols"]
        )

    def test_complexes_no_total_match(self, complexes: pd.DataFrame):
        res = Complexes.complex_genes(
            ["bar", "baz", "bar"], complexes=complexes, total_match=False
        )

        assert res.shape == (3, 2)
        assert all(
            any(v in ("bar", "baz") for v in vs.split("_"))
            for vs in res["components_genesymbols"]
        )


class TestAnnotations:
    def test_too_many_proteins_requested(self):
        with pytest.raises(
            ValueError,
            match=r"Please specify `force_full_download=True` in order to download the full dataset.",
        ):
            Annotations.get()

    def test_params(self):
        params = Annotations.params()
        assert Key.ORGANISM.value not in params

    def test_genesymbols_matter(self, cache_backup, requests_mock, tsv_data: bytes):
        url = urljoin(options.url, Annotations._query_type.endpoint)
        requests_mock.register_uri(
            "GET", f"{url}?proteins=bar&genesymbols=1&format=tsv", content=tsv_data
        )
        df = pd.read_csv(StringIO(tsv_data.decode("utf-8")), sep="\t")

        res = Annotations.get(proteins=["bar"], genesymbols=True)

        np.testing.assert_array_equal(res.index, df.index)
        np.testing.assert_array_equal(res.columns, df.columns)
        np.testing.assert_array_equal(res.values, df.values)

    def test_invalid_organism_does_not_matter(
        self, cache_backup, requests_mock, tsv_data: bytes
    ):
        url = urljoin(options.url, Annotations._query_type.endpoint)
        requests_mock.register_uri(
            "GET", f"{url}?proteins=foo&format=tsv", content=tsv_data
        )
        df = pd.read_csv(StringIO(tsv_data.decode("utf-8")), sep="\t")

        res = Annotations.get(proteins=["foo", "foo"], organism="foobarbaz")

        np.testing.assert_array_equal(res.index, df.index)
        np.testing.assert_array_equal(res.columns, df.columns)
        np.testing.assert_array_equal(res.values, df.values)

    @pytest.mark.parametrize(
        "n_prots",
        [
            _MAX_N_PROTS - 1,
            _MAX_N_PROTS,
            _MAX_N_PROTS + 1,
            2 * _MAX_N_PROTS,
            3 * _MAX_N_PROTS + 42,
        ],
    )
    def test_downloading_more_than_n_proteins(
        self, n_prots: int, cache_backup, requests_mock, tsv_data: bytes
    ):
        url = urljoin(options.url, Annotations._query_type.endpoint)
        prots = sorted(f"foo{i}" for i in range(n_prots))

        for i in range((n_prots // _MAX_N_PROTS) + 1):
            tmp = prots[i * _MAX_N_PROTS : (i + 1) * _MAX_N_PROTS]
            if not len(tmp):
                break
            requests_mock.register_uri(
                "GET",
                f"{url}?format=tsv&proteins={'%2C'.join(tmp)}",
                content=tsv_data,
            )
        df = pd.read_csv(StringIO(tsv_data.decode("utf-8")), sep="\t")

        res = Annotations.get(proteins=prots)

        np.testing.assert_array_equal(res.columns, df.columns)
        assert len(res) == (len(df) * int(np.ceil(n_prots / _MAX_N_PROTS)))

        if n_prots <= _MAX_N_PROTS:
            np.testing.assert_array_equal(res.index, df.index)
            np.testing.assert_array_equal(res.values, df.values)

    def test_resources(self, cache_backup, requests_mock, tsv_data: bytes):
        url = urljoin(options.url, Annotations._query_type.endpoint)
        requests_mock.register_uri(
            "GET", f"{url}?format=tsv&proteins=foo&resources=bar", content=tsv_data
        )
        df = pd.read_csv(StringIO(tsv_data.decode("utf-8")), sep="\t")

        res = Annotations.get("foo", resources="bar")

        np.testing.assert_array_equal(res.index, df.index)
        np.testing.assert_array_equal(res.columns, df.columns)
        np.testing.assert_array_equal(res.values, df.values)


class TestSignedPTMs:
    def test_get_signed_ptms_wrong_ptms_type(self):
        with pytest.raises(TypeError, match=r"Expected `ptms`"):
            SignedPTMs.get(42, pd.DataFrame())

    def test_get_signed_ptms_wrong_interactions_type(self):
        with pytest.raises(TypeError, match=r"Expected `interactions`"):
            SignedPTMs.get(pd.DataFrame(), 42)

    def test_get_signed_ptms(self):
        ptms = pd.DataFrame(
            {"enzyme": ["alpha", "beta", "gamma"], "substrate": [0, 1, 0], "foo": 42}
        )
        interactions = pd.DataFrame(
            {
                "source": ["gamma", "beta", "delta"],
                "target": [0, 0, 1],
                "is_stimulation": True,
                "is_inhibition": False,
                "bar": 1337,
            }
        )
        expected = pd.merge(
            ptms,
            interactions[["source", "target", "is_stimulation", "is_inhibition"]],
            left_on=["enzyme", "substrate"],
            right_on=["source", "target"],
            how="left",
        )

        res = SignedPTMs.get(ptms, interactions)

        np.testing.assert_array_equal(res.index, expected.index)
        np.testing.assert_array_equal(res.columns, expected.columns)

        np.testing.assert_array_equal(pd.isnull(res), pd.isnull(expected))
        np.testing.assert_array_equal(
            res.values[~pd.isnull(res)], expected.values[~pd.isnull(expected)]
        )

    def test_graph_not_a_dataframe(self):
        with pytest.raises(
            TypeError, match=r"Expected `data` to be of type `pandas.DataFrame`,"
        ):
            SignedPTMs.graph(42)

    def test_graph_empty_ptms(self):
        ptms = pd.DataFrame()
        interactions = pd.DataFrame(
            {
                "source": ["gamma", "beta", "delta"],
                "target": [0, 0, 1],
                "is_stimulation": True,
                "is_inhibition": False,
                "bar": 1337,
            }
        )

        with pytest.raises(ValueError, match=r"No PTMs were retrieved. Please"):
            SignedPTMs.get(ptms, interactions)

    def test_graph_empty_interactions(self):
        ptms = pd.DataFrame(
            {
                "enzyme": ["alpha", "beta", "gamma"],
                "substrate": [0, 1, 0],
                "foo": 42,
                "enzyme_genesymbol": "bar",
                "substrate_genesymbol": "baz",
            }
        )
        interactions = pd.DataFrame()

        with pytest.raises(ValueError, match=r"No interactions were retrieved. Please"):
            SignedPTMs.get(ptms, interactions)

    def test_graph_source_target(self):
        ptms = pd.DataFrame(
            {
                "enzyme": ["alpha", "beta", "gamma"],
                "substrate": [0, 1, 0],
                "foo": 42,
                "enzyme_genesymbol": "bar",
                "substrate_genesymbol": "baz",
            }
        )
        src, tgt = SignedPTMs._get_source_target_cols(ptms)

        assert src == "enzyme_genesymbol"
        assert tgt == "substrate_genesymbol"

        src, tgt = SignedPTMs._get_source_target_cols(
            ptms[ptms.columns.difference(["enzyme_genesymbol", "substrate_genesymbol"])]
        )

        assert src == "enzyme"
        assert tgt == "substrate"

    def test_graph(self):
        import networkx as nx

        ptms = pd.DataFrame(
            {
                "enzyme": ["alpha", "beta", "gamma"],
                "substrate": [0, 1, 0],
                "foo": 42,
                "references": "bar;baz",
            }
        )
        interactions = pd.DataFrame(
            {
                "source": ["gamma", "beta", "delta"],
                "target": [0, 0, 1],
                "is_stimulation": True,
                "is_inhibition": False,
                "bar": 1337,
            }
        )
        expected = pd.merge(
            ptms,
            interactions[["source", "target", "is_stimulation", "is_inhibition"]],
            left_on=["enzyme", "substrate"],
            right_on=["source", "target"],
            how="left",
        )

        G = SignedPTMs.graph(expected)

        assert isinstance(G, nx.DiGraph)
        assert G.is_directed()
        assert len(G) == 5
        for _, _, attr in G.edges(data=True):
            assert attr["foo"] == 42
            assert attr["references"] == ["bar", "baz"]


class TestUtils:
    def test_split_unique_join_no_func(self, string_series: pd.Series):
        res = _split_unique_join(string_series)

        np.testing.assert_array_equal(
            res, pd.Series(["foo:123", "bar:45;baz", None, "bar:67;baz:67", "foo"])
        )

    def test_split_unique_join_func(self, string_series: pd.Series):
        res = _split_unique_join(string_series, func=len)

        np.testing.assert_array_equal(res, pd.Series([1, 2, None, 2, 3], dtype=object))

    def test_strip_resource_label_no_func(self, string_series: pd.Series):
        res = _strip_resource_label(string_series, func=None)

        np.testing.assert_array_equal(
            res, pd.Series(["123", "45;baz", None, "67", "foo"])
        )

    def test_strip_resource_label_func(self):
        res = _strip_resource_label(
            pd.Series(["abc:123;bcd:123", "aaa:123", "a:1;b:2;c:3"]),
            func=lambda row: len(set(row)),
        )

        np.testing.assert_array_equal(res, pd.Series([1, 1, 3]))



from abc import ABC, ABCMeta
from enum import Enum, EnumMeta, unique
from typing import Any, Callable
from functools import wraps

from enum_tools import document_enum
import enum_tools.documentation as documentation

documentation.INTERACTIVE = True


def _pretty_raise_enum(cls: EnumMeta, fun: Callable) -> Callable:
    @wraps(fun)
    def wrapper(*args, **kwargs) -> Enum:
        try:
            return fun(*args, **kwargs)
        except ValueError as e:
            _cls, value, *_ = args
            e.args = (cls._format(value),)
            raise e

    if not issubclass(cls, ErrorFormatter):
        raise TypeError(f"Class `{cls}` must be subtype of `ErrorFormatter`.")
    elif not len(cls.__members__):
        # empty enum, for class hierarchy
        return fun

    return wrapper


class NoValue(Enum):
    """Enumeration which hides its :paramref:`value`."""

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.name}>"


class ErrorFormatter(ABC):  # noqa: D101
    __error_format__ = "Invalid value `{}` for `{}`. Valid options are: `{}`."

    @classmethod
    def _format(cls, value: Any) -> str:
        """Format the error message for invalid ``value``."""
        return cls.__error_format__.format(
            value, cls.__name__, [m.value for m in cls.__members__.values()]
        )


class FormatterMeta(EnumMeta, ABCMeta):  # noqa: D101
    def __call__(cls, *args, **kw):  # noqa: D102
        if getattr(cls, "__error_format__", None) is None:
            raise TypeError(
                f"Can't instantiate class `{cls.__name__}` "
                f"without `__error_format__` class attribute."
            )
        return super().__call__(*args, **kw)

    def __new__(cls, clsname, superclasses, attributedict):  # noqa: D102
        res = super().__new__(cls, clsname, superclasses, attributedict)
        res.__new__ = _pretty_raise_enum(res, res.__new__)
        return res


class PrettyEnumMixin(ErrorFormatter, NoValue, metaclass=FormatterMeta):
    """Enum mixin that that pretty prints when user uses invalid value."""

    @property
    def s(self) -> str:
        """Return the :paramref:`value` as :class:`str`."""
        return str(self.value)


@unique
@document_enum
class License(PrettyEnumMixin):
    """License types."""

    ACADEMIC = "academic"  # doc: Academic license.
    COMMERCIAL = "commercial"  # doc: Commercial license.
    NON_PROFIT = "non_profit"  # doc: Non-profit license.
    FOR_PROFIT = "for_profit"  # doc: For-profit license.
    IGNORE = "ignore"  # doc: Ignore the license type.


@unique
class InteractionDataset(PrettyEnumMixin):
    """Available interaction datasets in [OmniPath]_."""

    DOROTHEA = "dorothea"
    KINASE_EXTRA = "kinaseextra"
    LIGREC_EXTRA = "ligrecextra"
    LNCRNA_MRNA = "lncrna_mrna"
    MIRNA_TARGET = "mirnatarget"
    OMNIPATH = "omnipath"
    PATHWAY_EXTRA = "pathwayextra"
    TF_MIRNA = "tf_mirna"
    TF_REGULONS = "tfregulons"
    TF_TARGET = "tf_target"


@unique
@document_enum
class Organism(PrettyEnumMixin):
    """Organism types."""

    HUMAN = "human"  # doc: NCIB taxonomy id: 9606.
    MOUSE = "mouse"  # doc: NCIB taxonomy id: 10090.
    RAT = "rat"  # doc: NCIB taxonomy id: 10116.

    def __new__(cls, value: str):  # noqa: D102
        obj = object.__new__(cls)
        obj._code = {"human": 9606, "rat": 10116, "mouse": 10090}[value]
        return obj

    @property
    def code(self) -> int:
        """Return the code of this organism."""
        return self._code


__all__ = [
    License,
    Organism,
    InteractionDataset,
]



from io import BytesIO
from abc import ABC, ABCMeta, abstractmethod
from enum import Enum
from typing import (
    Any,
    Dict,
    Tuple,
    Union,
    Mapping,
    Callable,
    Iterable,
    Optional,
    Sequence,
)
from operator import itemgetter
from functools import wraps, partial
import logging

from pandas.api.types import is_float_dtype, is_numeric_dtype
import pandas as pd

from omnipath import options
from omnipath.constants import License, Organism
from omnipath._core.query import QueryType
from omnipath._core.utils._docs import d
from omnipath._core.requests._utils import (
    _ERROR_EMPTY_FMT,
    _inject_params,
    _inject_api_method,
    _strip_resource_label,
)
from omnipath.constants._pkg_constants import DEFAULT_FIELD, Key, Format, final
from omnipath._core.downloader._downloader import Downloader


def _error_handler(callback: Callable[[BytesIO], Any]) -> Callable:
    @wraps(callback)
    def wrapper(cls, *args, **kwargs) -> pd.DataFrame:
        res: pd.DataFrame = callback(*args, **kwargs)
        if len(res.columns) == 1 and res.columns == ["Something is not entirely good:"]:
            raise RuntimeError(" ".join(res.iloc[:, 0]))

        return res

    return wrapper


class OmnipathRequestMeta(ABCMeta):  # noqa: D101
    def __new__(cls, clsname, superclasses, attributedict):  # noqa: D102
        for supercls in superclasses:
            for attr in ("__string__", "__logical__", "__categorical__"):
                attributedict[attr] = attributedict.get(attr, frozenset()) | getattr(
                    supercls, attr, frozenset()
                )

        clazz = super().__new__(cls, clsname, superclasses, attributedict)
        _inject_api_method(clazz)

        return clazz


class OmnipathRequestABC(ABC, metaclass=OmnipathRequestMeta):
    """Base class for all :mod:`omnipath` requests."""

    __string__ = frozenset({"uniprot", "genesymbol"})
    __logical__ = frozenset()
    __categorical__ = frozenset()

    _json_reader = _error_handler(partial(pd.read_json, typ="frame"))
    _tsv_reader = _error_handler(
        partial(pd.read_csv, sep="\t", header=0, squeeze=False)
    )
    _query_type: Optional[QueryType] = None

    def __init__(self):
        self._downloader = Downloader(options)

    @classmethod
    @d.dedent
    def resources(cls, **kwargs) -> Tuple[str]:
        """%(query_resources)s"""
        return cls()._resources(**kwargs)

    @classmethod
    @d.dedent
    def params(cls) -> Dict[str, Any]:
        """%(query_params)s"""
        return {q.param: q.valid for q in cls._query_type.value}

    @classmethod
    def _annotations(cls) -> Dict[str, type]:
        """Return the type annotation for the query parameters."""
        return {q.param: q.annotation for q in cls._query_type.value}

    @classmethod
    def _docs(cls) -> Dict[str, Optional[str]]:
        """Return the type annotation for the query parameters."""
        return {q.param: q.doc for q in cls._query_type.value}

    def _get(self, **kwargs) -> pd.DataFrame:
        kwargs = self._modify_params(kwargs)
        kwargs = self._inject_fields(kwargs)
        kwargs, callback = self._convert_params(kwargs)
        kwargs = self._validate_params(kwargs)
        kwargs = self._finalize_params(kwargs)

        res = self._downloader.maybe_download(
            self._query_type.endpoint, params=kwargs, callback=callback, is_final=False
        )

        if self._downloader._options.convert_dtypes:
            res = self._convert_dtypes(res)

        return self._post_process(res)

    def _convert_params(
        self, params: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], Callable]:
        organism = params.pop("organism", params.pop("organisms", None))
        if organism is not None:
            organism = Organism(organism)
            try:
                params[self._query_type("organism").param] = organism.code
            except ValueError:
                pass

        # check the requested format
        fmt = params.pop("format", params.pop("formats", None))
        fmt = Format(Format.TSV if fmt is None else fmt)
        if fmt not in (Format.TSV, Format.JSON):
            logging.warning(
                f"Invalid `{Key.FORMAT.s}={fmt.s!r}`. Using `{Key.FORMAT.s}={Format.TSV.s!r}`"
            )
            fmt = Format.TSV
        callback = self._tsv_reader if fmt == Format.TSV else self._json_reader
        try:
            params[self._query_type("format").param] = fmt.s
        except ValueError:
            pass

        # check the license
        license = params.pop(
            "license", params.pop("licenses", self._downloader._options.license)
        )
        if license is not None:
            license = License(license)
            try:
                params[self._query_type("license").param] = license
            except ValueError:
                pass

        if self._downloader._options.password is not None:
            params.setdefault(Key.PASSWORD.s, self._downloader._options.password)

        return params, callback

    def _inject_fields(self, params: Dict[str, Any]) -> Dict[str, Any]:
        try:
            _inject_params(
                params,
                key=self._query_type(Key.FIELDS.value).param,
                value=getattr(DEFAULT_FIELD, self._query_type.name).value,
            )
        except AttributeError:
            # no default field for this query
            pass
        except Exception as e:
            logging.warning(
                f"Unable to inject `{Key.FIELDS.value}` for `{self}`. Reason: `{e}`"
            )

        return params

    def _validate_params(
        self, params: Dict[str, Any]
    ) -> Dict[str, Optional[Union[str, Sequence[str]]]]:
        """For each passed parameter, validate if it has the correct value."""
        res = {}
        for k, v in params.items():
            # first get the validator for the parameter, then validate
            res[self._query_type(k).param] = self._query_type(k)(v)
        return res

    def _finalize_params(self, params: Dict[str, Any]) -> Dict[str, str]:
        """Convert all the parameters to strings."""
        # this is largely redundant
        res = {}
        for k, v in params.items():
            if isinstance(v, str):
                res[k] = v
            elif isinstance(v, bool):
                res[k] = str(int(v))
            elif isinstance(v, (int, float)):
                res[k] = str(v)
            elif isinstance(v, Iterable):
                res[k] = ",".join(sorted(v))
            elif isinstance(v, Enum):
                res[k] = str(v.value)
            elif v is not None:
                logging.warning(f"Unable to process parameter `{k}={v}`. Ignoring")

        return dict(sorted(res.items(), key=itemgetter(0)))

    def _convert_dtypes(self, res: pd.DataFrame, **_) -> pd.DataFrame:
        """Automatically convert dtypes for this type of query."""

        def to_logical(col: pd.Series) -> pd.Series:
            if is_numeric_dtype(col):
                return col > 0
            return col.astype(str).str.lower().isin(("y", "t", "yes", "true", "1"))

        def handle_logical(df: pd.DataFrame, columns: frozenset) -> None:
            cols = list(frozenset(df.columns) & columns)
            if cols:
                df[cols] = df[cols].apply(to_logical)

        def handle_categorical(df: pd.DataFrame, columns: frozenset) -> None:
            cols = frozenset(df.columns) & columns
            cols = [
                col
                for col, dtype in zip(cols, df[cols].dtypes)
                if not is_float_dtype(dtype)
            ]
            if cols:
                df[cols] = df[cols].astype("category")

        def handle_string(df: pd.DataFrame, columns: frozenset) -> None:
            for col in frozenset(df.columns) & columns:
                mask = pd.isnull(df[col])
                df[col] = df[col].astype(str)
                df.loc[mask, col] = None

        if not isinstance(res, pd.DataFrame):
            raise TypeError(
                f"Expected the result to be of type `pandas.DataFrame`, found `{type(res).__name__}`."
            )

        handle_logical(res, self.__logical__)
        handle_categorical(res, self.__categorical__)
        handle_string(res, self.__string__)

        return res

    def _resources(self, **kwargs) -> Tuple[str]:
        """
        Return available resources for this type of query.

        Parameters
        ----------
        **kwargs
            Keyword arguments used for filtering unwanted resources.

        Returns
        -------
        tuple
            Unique and sorted resources.
        """
        return tuple(
            sorted(
                res
                for res, params in self._downloader.resources.items()
                if self._query_type.endpoint in params.get(Key.QUERIES.s, {})
                and self._resource_filter(
                    params[Key.QUERIES.s][self._query_type.endpoint], **kwargs
                )
            )
        )

    def _modify_params(self, params: Dict[str, Any]) -> Dict[str, Any]:
        """
        Remove parameters from this query.

        Parameters
        ----------
        params
            The parameters to filter.

        Returns
        -------
        :class:`dict`
            The filtered parameters.
        """
        return params

    @abstractmethod
    def _post_process(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Post process the result, e.g. by adding extra columns.

        df
            The result from :meth:`get`.

        Returns
        -------
        :class:`pandas.DataFrame`
            The maybe modified result.
        """
        pass

    @abstractmethod
    def _resource_filter(self, data: Mapping[str, Any], **kwargs) -> bool:
        """
        Filter out resources relevant to this query.

        Parameters
        ----------
        data
            Data which is used as a basis for the filtering.
        kwargs
            Additional keyword arguments.

        Returns
        --------
        bool
            `True` if the resource should be included, otherwise `False`.
        """
        pass

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}>"

    def __repr__(self) -> str:
        return str(self)


class CommonPostProcessor(OmnipathRequestABC, ABC):
    """
    Class that implements common post-processing steps for :class:`omnipath.requests.Enzsub`, \
    :class:`omnipath.requests.Intercell`, :class:`omnipath.requests.Complexes` and \
    :class`omnipath.interactions.InteractionRequest`.

    This class remove `'genesymbols'` and `'organisms'` from the query parameters, as well as optionally adds
    number of resources and references to the result or removes the resource labels from references (PubMed IDs)
    :class:`omnipath.interactions.InteractionRequest` and :class:`omnipath.requests.Enzsub`.
    """

    def _post_process(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Add number of resources and references for each row in the resulting ``df``.

        Parameters
        ----------
        df
            The dataframe containing results.

        Returns
        -------
        :class:`pandas.DataFrame`
            The modified dataframe.
        """
        if "references" in df:
            df["references_stripped"] = _strip_resource_label(df["references"])
            df["n_references"] = _strip_resource_label(
                df["references"], func=lambda row: len(set(row))
            )
        if "sources" in df:
            df["n_sources"] = df["sources"].astype(str).str.split(";").apply(len)
            df["n_primary_sources"] = (
                df["sources"]
                .astype(str)
                .str.split(";")
                .apply(
                    lambda row: len(
                        [r for r in row if "_" not in r]
                        if isinstance(row, Iterable)
                        else 0
                    )
                )
            )

        return df


class OrganismGenesymbolsRemover(CommonPostProcessor, ABC):
    """Class that removes organism and genesymbols keys from the query."""

    def _modify_params(self, params: Dict[str, Any]) -> Dict[str, Any]:
        params.pop(Key.ORGANISM.s, None)
        params.pop(Key.GENESYMBOLS.s, None)

        return params

    @classmethod
    @d.dedent
    def params(cls) -> Dict[str, Any]:
        """%(query_params)s"""
        params = super().params()
        params.pop(Key.ORGANISM.s, None)
        params.pop(Key.GENESYMBOLS.s, None)

        return params


class GraphLike(ABC):
    """
    Class that is able to construct a graph.

    Should be injected with any class with :meth:`get`.
    """

    @classmethod
    @abstractmethod
    def _get_source_target_cols(cls, data: pd.DataFrame) -> Tuple[str, str]:
        pass

    @classmethod
    def graph(cls, data: Optional[pd.DataFrame] = None, **kwargs):
        """
        Create a graph.

        Parameters
        ----------
        data
            The interaction data. If `None`, create a new request.
        kwargs
            Keyword arguments for :meth:`get` if ``data = None``.

        Returns
        -------
        :class:`networkx.DiGraph`
            The interaction graph.
        """
        try:
            import networkx as nx
        except ImportError:
            raise ImportError(
                "Unable to import `networkx`. Please install it as `pip install network`."
            ) from None

        data = cls.get(**kwargs) if data is None else data

        if not isinstance(data, pd.DataFrame):
            raise TypeError(
                f"Expected `data` to be of type `pandas.DataFrame`, found `{type(data).__name__}`."
            )
        if data.empty:
            raise ValueError(_ERROR_EMPTY_FMT.format(obj="data"))

        source, target = cls._get_source_target_cols(data)
        G = nx.from_pandas_edgelist(
            data,
            source=source,
            target=target,
            edge_attr=tuple(data.columns.difference([source, target])),
            create_using=nx.DiGraph,
        )

        for s, t, attr in G.edges(data=True):
            for col in ["references", "references_stripped", "sources"]:
                if col in data:
                    if ";" in str(attr[col]):
                        G.edges[s, t][col] = sorted(str(attr[col]).split(";"))

        return G


class Enzsub(CommonPostProcessor):
    """
    Request enzyme-substrate relationships from [OmniPath]_.

    Imports the enzyme-substrate (more exactly, enzyme-PTM) relationships `database <https://omnipathdb.org/enzsub>`__.
    """

    __string__ = frozenset({"enzyme", "substrate"})
    __categorical__ = frozenset({"residue_type", "modification"})

    _query_type = QueryType.ENZSUB

    def _resource_filter(self, data: Mapping[str, Any], **_) -> bool:
        return True


@final
class SignedPTMs(Enzsub, GraphLike):
    """
    Request enzyme-substrate relationships and interactions from [OmniPath]_.

    PTM data does not contain sign (activation/inhibition), we generate this information based on the
    interaction network.
    """

    @classmethod
    def _get_source_target_cols(cls, data: pd.DataFrame) -> Tuple[str, str]:
        source = "enzyme_genesymbol" if "enzyme_genesymbol" in data else "enzyme"
        target = (
            "substrate_genesymbol" if "substrate_genesymbol" in data else "substrate"
        )

        return source, target

    @classmethod
    def get(
        cls,
        ptms: Optional[pd.DataFrame] = None,
        interactions: Optional[pd.DataFrame] = None,
    ) -> pd.DataFrame:
        """
        Get signs for enzyme-PTM interactions.

        Parameters
        ----------
        ptms
            Data generated by :meth:`omnipath.requests.Enzsub.get`. If `None`, a new request will be performed.
        interactions
            Data generated by :meth:`omnipath.interactions.OmniPath.get`.  If `None`, a new request will be performed.

        Returns
        -------
        :class:`pandas.DataFrame`
            The signed PTMs with columns **'is_inhibition'** and **'is_stimulation'**.
        """
        from omnipath.requests import Enzsub
        from omnipath.interactions import OmniPath

        ptms = Enzsub.get() if ptms is None else ptms
        interactions = OmniPath.get() if interactions is None else interactions

        if not isinstance(ptms, pd.DataFrame):
            raise TypeError(
                f"Expected `ptms` to be of type `pandas.DataFrame`, found `{type(ptms).__name__}`."
            )
        if not isinstance(interactions, pd.DataFrame):
            raise TypeError(
                f"Expected `interactions` to be of type `pandas.DataFrame`, found `{type(ptms).__name__}`."
            )
        if ptms.empty:
            raise ValueError(_ERROR_EMPTY_FMT.format(obj="PTMs"))
        if interactions.empty:
            raise ValueError(_ERROR_EMPTY_FMT.format(obj="interactions"))

        return pd.merge(
            ptms,
            interactions[["source", "target", "is_stimulation", "is_inhibition"]],
            left_on=["enzyme", "substrate"],
            right_on=["source", "target"],
            how="left",
        )


__all__ = [Enzsub, SignedPTMs]



from abc import ABC, abstractmethod
from shutil import rmtree
from typing import Any, Union, Optional
from pathlib import Path
import os
import pickle


class Cache(ABC):
    """Abstract class which defines the caching interface."""

    @abstractmethod
    def __getitem__(self, key: str) -> Optional[Any]:
        pass

    @abstractmethod
    def __setitem__(self, key: str, value: Any) -> None:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def clear(self) -> None:  # noqa: D102
        pass

    @property
    @abstractmethod
    def path(self) -> Optional[Union[str, Path]]:  # noqa: D102
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    def __repr__(self) -> str:
        return str(self)


class FileCache(Cache):
    """
    Cache which persists the data into :mod:`pickle` files.

    Parameters
    ----------
    path
        Path to a directory where the files will be stored.
    """

    _suffix = ".pickle"

    def __init__(self, path: Union[str, Path]):
        if not isinstance(path, (str, Path)):
            raise TypeError(
                f"Expected `path` to be either `str` or `pathlib.Path`, "
                f"found `{type(path).__name__}`."
            )
        if not str(path):
            raise ValueError("Empty cache path.")

        self._cache_dir = Path(path)

    def __contains__(self, key: str) -> bool:
        if not key.endswith(self._suffix):
            key += self._suffix

        return (self._cache_dir / key).is_file()

    def __setitem__(self, key: str, value: Any):
        self._cache_dir.mkdir(parents=True, exist_ok=True)

        fname = str(key)
        if not fname.endswith(self._suffix):
            fname += self._suffix

        with open(self._cache_dir / fname, "wb") as fout:
            pickle.dump(value, fout)

    def __getitem__(self, key: str) -> Any:
        if not key.endswith(self._suffix):
            key += self._suffix

        if not (self._cache_dir / key).is_file():
            raise KeyError(self._cache_dir / key)

        with open(self._cache_dir / key, "rb") as fin:
            return pickle.load(fin)

    def __len__(self) -> int:
        return (
            len([f for f in os.listdir(self.path) if str(f).endswith(self._suffix)])
            if self.path.is_dir()
            else 0
        )

    @property
    def path(self) -> Path:
        """Return the directory where the cache files are stored."""
        return self._cache_dir

    def clear(self) -> None:
        """Remove all files and the directory under :paramref:`path`."""
        if self._cache_dir.is_dir():
            rmtree(self._cache_dir)

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}[size={len(self)}, path={str(self.path)!r}]>"


class MemoryCache(dict, Cache):
    """Cache which persists the data into the memory."""

    @property
    def path(self) -> None:
        """Return `None`."""
        return None

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}[size={len(self)}]>"

    def __repr__(self) -> str:
        return str(self)

    def __copy__(self) -> "MemoryCache":
        return self

    def copy(self) -> "MemoryCache":
        """Return self."""
        return self


def clear_cache() -> None:
    """Remove all cached data from :attr:`omnipath.options.cache`."""
    from omnipath import options

    options.cache.clear()


__all__ = [clear_cache]



from copy import copy, deepcopy
from pathlib import Path

import pytest

from omnipath import options, clear_cache
from omnipath._core.cache._cache import FileCache, MemoryCache


def test_clear_cache_high_lvl(cache_backup):
    options.cache["foo"] = 42
    assert len(options.cache) == 1
    assert options.cache["foo"] == 42

    clear_cache()

    assert len(options.cache) == 0


class TestMemoryCache:
    def test_str_repr(self):
        mc = MemoryCache()

        assert str(mc) == f"<{mc.__class__.__name__}[size={len(mc)}]>"
        assert repr(mc) == f"<{mc.__class__.__name__}[size={len(mc)}]>"

    def test_path_is_None(self):
        mc = MemoryCache()
        assert mc.path is None

    def test_copy_does_nothing(self):
        mc = MemoryCache()

        assert mc is mc.copy()
        assert mc is copy(mc)

    def test_deepcopy_work(self):
        mc = MemoryCache()

        assert mc is not deepcopy(mc)

    def test_cache_works(self):
        mc = MemoryCache()
        sentinel = object()

        mc["foo"] = sentinel

        assert len(mc) == 1
        assert mc["foo"] is sentinel

        mc.clear()

        assert len(mc) == 0


class TestPickleCache:
    def test_invalid_path(self):
        with pytest.raises(TypeError):
            FileCache(42)

    def test_path(self, tmpdir):
        fc = FileCache(Path(tmpdir))

        assert isinstance(fc.path, Path)
        assert str(fc.path) == str(tmpdir)

    def test_str_repr(self, tmpdir):
        fc = FileCache(Path(tmpdir))

        assert (
            str(fc)
            == f"<{fc.__class__.__name__}[size={len(fc)}, path={str(tmpdir)!r}]>"
        )
        assert (
            repr(fc)
            == f"<{fc.__class__.__name__}[size={len(fc)}, path={str(tmpdir)!r}]>"
        )

    def test_cache_works(self, tmpdir):
        fc = FileCache(Path(tmpdir))
        sentinel = object()

        assert "foo" not in fc
        fc["foo"] = 42
        fc["bar.pickle"] = sentinel

        assert "foo" in fc
        assert "foo.pickle" in fc
        assert fc["bar.pickle"] is not sentinel

    def test_clear_works(self, tmpdir):
        fc = FileCache(Path(tmpdir))
        fc["foo"] = 42
        assert Path(fc.path).exists()

        fc.clear()

        assert len(fc) == 0
        assert not Path(tmpdir).exists()



from os import remove
from typing import Optional
from pathlib import Path
from configparser import NoSectionError

import pytest

from omnipath.constants import License
from omnipath._core.utils._options import Options
from omnipath.constants._pkg_constants import DEFAULT_OPTIONS


class TestOptions:
    def test_invalid_url_type(self, options: Options):
        with pytest.raises(TypeError):
            options.url = 42

    def test_invalid_url(self, options: Options):
        with pytest.raises(ValueError):
            options.url = "foo"

    def test_invalid_license(self, options: Options):
        with pytest.raises(ValueError):
            options.license = "foo"

    def test_invalid_cache_type(self, options: Options):
        with pytest.raises(TypeError):
            options.cache = 42

    def test_invalid_password_type(self, options: Options):
        with pytest.raises(TypeError):
            options.password = 42

    def test_invalid_num_retries(self, options: Options):
        with pytest.raises(ValueError):
            options.num_retries = -1

    def test_invalid_timeout(self, options: Options):
        with pytest.raises(ValueError):
            options.timeout = 0

    def test_invalid_chunk_size(self, options: Options):
        with pytest.raises(ValueError):
            options.chunk_size = 0

    def test_from_options_invalid_type(self):
        with pytest.raises(TypeError):
            Options.from_options("foo")

    def test_url_localhost(self, options: Options):
        options.url = "https://localhost"

        assert options.url == "https://localhost"

    @pytest.mark.parametrize("license", list(License))
    def test_valid_license(self, options: Options, license: License):
        options.license = license.value

        assert isinstance(options.license, License)
        assert options.license == license

    @pytest.mark.parametrize("pwd", ["foo", None])
    def test_password(self, options: Options, pwd: Optional[str]):
        options.password = pwd

        assert options.password == pwd

    def test_from_options(self, options: Options):
        new_opt = Options.from_options(options)

        for k, v in options.__dict__.items():
            assert getattr(new_opt, k) == v

    def test_from_options_new_values(self, options: Options):
        new_opt = Options.from_options(
            options, autoload=not options.autoload, num_retries=0
        )

        for k, v in options.__dict__.items():
            if k not in ("autoload", "num_retries"):
                assert getattr(new_opt, k) == v

        assert new_opt.autoload != options.autoload
        assert new_opt.num_retries == 0

    def test_from_config_no_file(self, config_backup):
        if Path(Options.config_path).exists():
            remove(Options.config_path)

        new_opt = Options.from_config()

        for k, v in DEFAULT_OPTIONS.__dict__.items():
            if hasattr(new_opt, k) and not k.startswith("_"):
                assert getattr(new_opt, k) == v

    def test_from_config_section_is_not_url(self):
        with pytest.raises(NoSectionError, match=r"No section: 'http://foo.bar'"):
            Options.from_config("http://foo.bar")

    def test_write_config(self, options: Options, config_backup):
        options.timeout = 1337
        options.license = License.COMMERCIAL
        options.password = "foobarbaz"
        options.write()

        new_opt = Options.from_config()
        for k, v in options.__dict__.items():
            if k == "cache":
                assert type(new_opt.cache) == type(options.cache)  # noqa: E721
            elif k == "password":
                # don't store the password in the file
                assert getattr(new_opt, k) is None
            elif k not in ("timeout", "license"):
                assert getattr(new_opt, k) == v

        assert new_opt.timeout == 1337
        assert new_opt.license == License.COMMERCIAL

    def test_write_new_section(self, options: Options, config_backup):
        options.timeout = 42
        options.write("https://foo.bar")

        new_opt = Options.from_config("https://foo.bar")
        assert options is not new_opt
        for k, v in options.__dict__.items():
            if k == "url":
                assert v == options.url
                assert new_opt.url == "https://foo.bar"
            elif k == "cache":
                assert type(new_opt.cache) == type(options.cache)  # noqa: E721
            else:
                assert getattr(new_opt, k) == v

    def test_write_new_section_not_url(self, options: Options, config_backup):
        with pytest.raises(ValueError, match=r"Invalid URL: `foobar`."):
            options.write("foobar")

    def test_contextmanager(self, options: Options):
        with options as new_opt:
            assert options is not new_opt
            for k, v in options.__dict__.items():
                if k == "cache":
                    assert type(new_opt.cache) == type(options.cache)  # noqa: E721
                else:
                    assert getattr(new_opt, k) == v



from abc import ABCMeta
from enum import Enum, EnumMeta
from typing import Set, Tuple, Union, Optional, Sequence, FrozenSet

from inflect import engine

from omnipath.constants._constants import FormatterMeta, ErrorFormatter
from omnipath._core.query._query_validator import (
    EnzsubValidator,
    ComplexesValidator,
    IntercellValidator,
    AnnotationsValidator,
    InteractionsValidator,
)

_engine = engine()


def _get_synonyms(key: str) -> Tuple[str]:
    """
    Create synonyms for ``key``.

    This function creates just 2 synonyms - the singular and plural case of ``key``.

    Parameters
    ----------
    key
        Key for which to create the synonyms.

    Returns
    -------
    :class:`tuple`
        Synonyms for ``key``. User will be able to use these submitting requests.
    """
    if not isinstance(key, str):
        raise TypeError(f"Expected a `str`, found `{type(key)}`.")

    singular = _engine.singular_noun(key)
    singular = singular if isinstance(singular, str) else key

    plural = _engine.plural_noun(singular)
    if not isinstance(plural, str):
        plural = key + "s" if not key.endswith("s") else key

    return tuple(sorted({singular, plural}))


class SynonymizerMeta(EnumMeta, ABCMeta):  # noqa: D101
    def __new__(cls, clsname, superclasses, attributedict):  # noqa: D102
        validator = attributedict.get("__validator__", None)

        if validator is None:
            return super().__new__(cls, clsname, superclasses, attributedict)

        for k in list(validator):
            k = str(k.name)
            for i, synonym in enumerate(_get_synonyms(k.lower())):
                attributedict[f"{k}_{i}"] = synonym

        return super().__new__(cls, clsname, superclasses, attributedict)


class QueryMeta(SynonymizerMeta, FormatterMeta):  # noqa: D101
    pass


class Query(ErrorFormatter, Enum, metaclass=QueryMeta):  # noqa: D101
    @property
    def _query_name(self) -> str:
        """Convert the synonym to an actual query parameter name."""
        return "_".join(self.name.split("_")[:-1])

    @property
    def _delegate(self):
        """Delegate the validation."""
        return getattr(self.__validator__, self._query_name)

    @property
    def param(self) -> str:
        """Get the parameter name as required by the server."""
        return self._query_name.lower()

    @property
    def valid(self) -> Optional[FrozenSet[str]]:
        """Return the set of valid values for :paramref:`param`."""
        return self._delegate.valid

    @property
    def annotation(self) -> type:
        """Return type annotations for :paramref:`param`."""
        return self._delegate.annotation

    @property
    def doc(self) -> Optional[str]:
        """Return the docstring for :paramref:`param`."""
        return self._delegate.doc

    def __call__(
        self, value: Optional[Union[str, Sequence[str]]]
    ) -> Optional[Set[str]]:
        """%(validate)s"""  # noqa: D401
        return self._delegate(value)


class EnzsubQuery(Query):  # noqa: D101
    __validator__ = EnzsubValidator


class InteractionsQuery(Query):  # noqa: D101
    __validator__ = InteractionsValidator


class ComplexesQuery(Query):  # noqa: D101
    __validator__ = ComplexesValidator


class AnnotationsQuery(Query):  # noqa: D101
    __validator__ = AnnotationsValidator


class IntercellQuery(Query):  # noqa: D101
    __validator__ = IntercellValidator


class QueryType(Enum):  # noqa: D101
    ENZSUB = EnzsubQuery
    INTERACTIONS = InteractionsQuery
    COMPLEXES = ComplexesQuery
    ANNOTATIONS = AnnotationsQuery
    INTERCELL = IntercellQuery

    def __call__(
        self, value: Optional[Union[str, Sequence[str]]]
    ) -> Optional[Set[str]]:
        """%(validate)s"""  # noqa: D401
        return self.value(value)

    @property
    def endpoint(self) -> str:
        """Get the API endpoint for this type of query."""
        return self.name.lower()


__all__ = [
    EnzsubQuery,
    InteractionsQuery,
    ComplexesQuery,
    AnnotationsQuery,
    IntercellQuery,
]



from typing import _GenericAlias
from collections import defaultdict

import pytest

from omnipath._core.query._query import (
    Query,
    QueryType,
    EnzsubQuery,
    ComplexesQuery,
    IntercellQuery,
    AnnotationsQuery,
    InteractionsQuery,
    _get_synonyms,
)
from omnipath._core.query._query_validator import (
    EnzsubValidator,
    ComplexesValidator,
    IntercellValidator,
    AnnotationsValidator,
    InteractionsValidator,
    _to_string_set,
)


class TestUtils:
    def test_get_synonyms_wrong_type(self):
        with pytest.raises(TypeError):
            _get_synonyms(42)

    def test_get_synonyms_from_s2p(self):
        res = _get_synonyms("cat")

        assert len(res) == 2
        assert res == ("cat", "cats")

    def test_get_synonyms_from_p2s(self):
        res = _get_synonyms("dogs")

        assert len(res) == 2
        assert res == ("dog", "dogs")

    def test_to_string_set_string(self):
        assert {"foo"} == _to_string_set("foo")

    def test_to_string_set_int(self):
        assert {"42"} == _to_string_set(42)

    def test_to_string_set_sequence(self):
        assert {"foo", "42"} == _to_string_set(["foo", 42])


class TestValidator:
    @pytest.mark.parametrize(
        "validator",
        [
            EnzsubValidator,
            InteractionsValidator,
            ComplexesValidator,
            AnnotationsValidator,
            IntercellValidator,
        ],
    )
    def test_validator_no_server_access(self, validator):
        for value in list(validator):
            v = validator(value)

            assert v.valid is None
            assert v.doc is None

            assert v(None) is None
            assert v("foo") == {"foo"}
            assert v(42) == {"42"}
            assert v(True) == {"1"}
            assert v(False) == {"0"}
            assert v(["foo", "foo"]) == {"foo"}
            assert v(["foo", 42]) == {"foo", "42"}
            assert v({"foo", "bar", "baz"}) == {"foo", "bar", "baz"}

            assert issubclass(type(v.annotation), (_GenericAlias, type))


class TestQuery:
    @pytest.mark.parametrize(
        "query,validator",
        zip(
            [
                EnzsubQuery,
                InteractionsQuery,
                ComplexesQuery,
                AnnotationsQuery,
                IntercellQuery,
            ],
            [
                EnzsubValidator,
                InteractionsValidator,
                ComplexesValidator,
                AnnotationsValidator,
                IntercellValidator,
            ],
        ),
    )
    def test_query_correct_validator(self, query, validator):
        assert query.__validator__ == validator

    def test_query_endpoint(self):
        for q in list(QueryType):
            q = QueryType(q)

            assert issubclass(q.value, Query)
            assert q.endpoint == q.name.lower()

    @pytest.mark.parametrize(
        "query,validator",
        zip(
            [
                EnzsubQuery,
                InteractionsQuery,
                ComplexesQuery,
                AnnotationsQuery,
                IntercellQuery,
            ],
            [
                EnzsubValidator,
                InteractionsValidator,
                ComplexesValidator,
                AnnotationsValidator,
                IntercellValidator,
            ],
        ),
    )
    def test_query_delegation(self, query, validator, mocker):
        call_spy = mocker.spy(validator, "__call__")

        qdb = query("databases")
        _ = qdb("foo")

        call_spy.assert_called_once_with(
            getattr(qdb.__validator__, qdb._query_name), "foo"
        )
        assert call_spy.spy_return == {"foo"}
        assert qdb.doc is None

        for attr in ("valid", "annotation", "doc"):
            m = mocker.patch.object(
                validator, attr, new_callable=mocker.PropertyMock, return_value="foo"
            )
            assert getattr(qdb, attr) == "foo"

            m.assert_called_once()

    @pytest.mark.parametrize(
        "query",
        [
            EnzsubQuery,
            InteractionsQuery,
            ComplexesQuery,
            AnnotationsQuery,
            IntercellQuery,
        ],
    )
    def test_query_synonym(self, query):
        mapper = defaultdict(list)
        for v in list(query):
            name = "_".join(v.name.split("_")[:-1])
            mapper[name].append(v.value)

        for vs in mapper.values():
            assert len(vs) == 2
            assert len({query(v).param for v in vs})



from typing import Any, Tuple, Mapping, Iterable, Optional, Sequence

import pandas as pd

from omnipath._core.query import QueryType
from omnipath._core.query._types import Strseq_t
from omnipath._core.requests._request import OrganismGenesymbolsRemover
from omnipath.constants._pkg_constants import Key, Format, final
from omnipath._core.query._query_validator import _to_string_set


@final
class Intercell(OrganismGenesymbolsRemover):
    """
    Request `intercell` annotations from [OmniPath]_.

    Imports the [OmniPath]_ inter-cellular communication role annotation
    `database <https://omnipathdb.org/intercell>`__.

    It provides information on the roles in inter-cellular signaling, e.g. if a protein is a ligand, a receptor,
    an extracellular matrix (ECM) component, etc.
    """

    __categorical__ = frozenset(
        {"category", "parent", "database", "scope", "aspect", "source", "entity_type"}
    )

    _query_type = QueryType.INTERCELL

    def _resource_filter(
        self,
        data: Mapping[str, Any],
        generic_categories: Optional[Sequence[str]] = None,
        **kwargs,
    ) -> bool:
        return generic_categories is None or _to_string_set(
            data.get(Key.GENERIC_CATEGORIES.s, set())
        ) & _to_string_set(generic_categories)

    @classmethod
    def resources(cls, generic_categories: Strseq_t = None) -> Tuple[str]:
        """
        Return the resources falling into the specified generic categories.

        Parameters
        ----------
        generic_categories
            For valid options, see :paramref:`generic_categories`.

        Returns
        -------
        tuple
            The filtered resources according to ``generic_categories``.
        """
        if generic_categories is None:
            return super().resources()

        if isinstance(generic_categories, str):
            generic_categories = (generic_categories,)
        if not isinstance(generic_categories, (Sequence, Iterable)):
            raise TypeError(
                f"Expected generic categories to be a `str` or an `Iterable`, "
                f"found `{type(generic_categories).__name__}`."
            )

        if not len(generic_categories):
            raise ValueError("No generic categories have been selected.")

        return super().resources(**{Key.GENERIC_CATEGORIES.s: generic_categories})

    @classmethod
    def categories(cls) -> Tuple[str]:
        """Return categories from the `intercell` database."""
        return cls()._get_metadata(Key.CATEGORY.s)

    @classmethod
    def generic_categories(cls) -> Tuple[str]:
        """Return generic categories from the `intercell` database."""
        return cls()._get_metadata(Key.PARENT.s)

    def _get_metadata(self, col: Optional[str]) -> Tuple[str]:
        """Return unique summary data from column ``col``."""
        metadata = self._downloader.maybe_download(
            Key.INTERCELL_SUMMARY.s,
            params={Key.FORMAT.s: Format.JSON.s},
            callback=self._json_reader,
            is_final=False,
        )

        if col not in metadata.columns:
            raise KeyError(f"Column `{col}` not found in `{list(metadata.columns)}`.")

        return tuple(sorted(pd.unique(metadata[col].astype(str))))


__all__ = [Intercell]



from omnipath._core.requests.interactions._utils import import_intercell_network
from omnipath._core.requests.interactions._interactions import (
    TFmiRNA,
    Dorothea,
    OmniPath,
    TFtarget,
    KinaseExtra,
    LigRecExtra,
    PathwayExtra,
    AllInteractions,
    Transcriptional,
    PostTranslational,
    miRNA,
    lncRNAmRNA,
)



from typing import Iterable, Optional

from conftest import RTester
import pytest

import numpy as np
import pandas as pd

from omnipath.constants import License, Organism
import omnipath as op

# in order to minimize server access, the tests are not parametrized
# and the resources are chosen so that minimal data required is transferred
# these tests will also run only on 1 job CI matrix, to further reduce the load
#
# note that these tests don't test whether `<query>.params()` returns the valid values
# this would require different Python interpreter invocation, since by design, during testing
# `omnipath.options.autoload` is disabled


def _assert_dataframes_equal(
    expected: pd.DataFrame,
    actual: pd.DataFrame,
    clazz: type = op.requests.Enzsub,
    remove_metadata: Optional[Iterable[str]] = None,
):
    assert isinstance(expected, pd.DataFrame)
    assert isinstance(actual, pd.DataFrame)

    # some small naming discrepancy
    actual.rename(columns={"n_primary_sources": "n_resources"}, inplace=True)
    # these are always present in our case
    if remove_metadata is None:
        remove_metadata = ["n_sources", "references_stripped"]
    for k in remove_metadata:
        if k in actual.columns:
            del actual[k]

    np.testing.assert_array_equal(expected.shape, actual.shape)
    # don't compare index since sometimes, it's not informative + differs across calls from OmnipathR
    col_order = sorted(expected.columns)
    np.testing.assert_array_equal(col_order, sorted(actual.columns))

    expected = clazz()._convert_dtypes(expected)

    # this way, we know which column fails
    for col in col_order:
        e, a = expected[col], actual[col]
        emask = ~(pd.isna(e).values | pd.isnull(a).values)
        amask = ~(pd.isna(e).values | pd.isnull(a).values)

        # print(f"Column `{col}`")
        np.testing.assert_array_equal(emask, amask)
        np.testing.assert_array_equal(e[emask], a[emask])


class TestEnzSub(RTester):
    def test_organism(self):
        organism = Organism.RAT
        expected = self.omnipathr.import_omnipath_enzsub(
            resources="DEPOD", genesymbols=False, organism=organism.code
        )
        actual = op.requests.Enzsub.get(
            resources="DEPOD", genesymbols=False, organism=organism
        )

        _assert_dataframes_equal(expected, actual)

    def test_resources(self):
        expected = self.omnipathr.import_omnipath_enzsub(
            resources="HPRD", genesymbols=True
        )
        actual = op.requests.Enzsub.get(resources="HPRD", genesymbols=True)

        _assert_dataframes_equal(expected, actual)

    def test_fields(self):
        fields = ["isoforms", "ncbi_tax_id"]
        expected = self.omnipathr.import_omnipath_enzsub(
            resources="DEPOD", genesymbols=True, fields=fields
        )
        actual = op.requests.Enzsub.get(
            resources="DEPOD", genesymbols=True, fields=fields
        )

        _assert_dataframes_equal(expected, actual)

    def test_license(self):
        license = License.COMMERCIAL
        expected = self.omnipathr.import_omnipath_enzsub(
            resources="DEPOD", genesymbols=True, license=license.value
        )
        actual = op.requests.Enzsub.get(
            resources="DEPOD", genesymbols=True, license=license
        )

        _assert_dataframes_equal(expected, actual)


class TestIntercell(RTester):
    def test_categories(self):
        expected = sorted(self.omnipathr.get_intercell_categories())
        actual = sorted(op.requests.Intercell.categories())

        np.testing.assert_array_equal(expected, actual)

    def test_generic_categories(self):
        expected = sorted(self.omnipathr.get_intercell_generic_categories())
        actual = sorted(op.requests.Intercell.generic_categories())

        np.testing.assert_array_equal(expected, actual)

    def test_normal_run(self):
        expected = self.omnipathr.import_omnipath_intercell(
            causality="transmitter", scope="specific", entity_types="protein"
        )
        actual = op.requests.Intercell.get(
            causality="transmitter", scope="specific", entity_types="protein"
        )

        _assert_dataframes_equal(expected, actual)


class TestComplexes(RTester):
    def test_complex_genes(self):
        genes = ["ITGB1", "RET"]
        expected = self.omnipathr.import_omnipath_complexes(resources="CellPhoneDB")
        actual = op.requests.Complexes.get(database="CellPhoneDB")

        _assert_dataframes_equal(
            expected,
            actual,
            remove_metadata=[
                "n_sources",
                "n_resources",
                "n_references",
                "references_stripped",
            ],
        )

        expected = self.omnipathr.get_complex_genes(genes, complexes=expected)
        actual = op.requests.Complexes.complex_genes(genes, complexes=actual)

        _assert_dataframes_equal(
            expected,
            actual,
            remove_metadata=[
                "n_sources",
                "n_resources",
                "n_references",
                "references_stripped",
            ],
        )


class TestAnnotations(RTester):
    def test_normal_run(self):
        proteins = ["ITGB1", "RET"]
        expected = self.omnipathr.import_omnipath_annotations(
            proteins=proteins, resources="Phobius", genesymbols=False
        )
        actual = op.requests.Annotations.get(
            proteins=proteins, databases="Phobius", genesymbols=False
        )

        _assert_dataframes_equal(expected, actual)


class TestInteractions(RTester):
    def test_tfregulons_levels(self):
        fields = ["tfregulons_level", "tfregulons_tfbs"]
        expected = self.omnipathr.import_tf_target_interactions(
            resources=["ABS"], fields=fields, genesymbols=False
        )
        actual = op.interactions.TFtarget.get(
            resources=["ABS"], fields=fields, genesymbols=False
        )

        _assert_dataframes_equal(expected, actual)

    def test_dorothea_levels(self):
        fields = ["dorothea_level"]
        expected = self.omnipathr.import_dorothea_interactions(
            resources=["ABS"], dorothea_levels="D", fields=fields, genesymbols=False
        )
        actual = op.interactions.Dorothea.get(
            resources=["ABS"], dorothea_levels="D", fields=fields, genesymbols=False
        )

        _assert_dataframes_equal(expected, actual)

    def test_omnipath(self):
        expected = self.omnipathr.import_omnipath_interactions(
            resources="CA1", genesymbols=False
        )
        actual = op.interactions.OmniPath.get(resource="CA1", genesymbols=False)

        _assert_dataframes_equal(expected, actual)


class TestUtils(RTester):
    @pytest.mark.skip(reason="TODO: different index order, ref. mismatch")
    def test_import_intercell_network(self):
        from rpy2.robjects import ListVector

        interactions_params = {"resources": "CellPhoneDB"}
        transmitter_params = {"categories": "ligand"}
        receiver_params = {"categories": "receptor"}

        expected = self.omnipathr.import_intercell_network(
            interactions_param=ListVector(list(interactions_params.items())),
            transmitter_param=ListVector(list(transmitter_params.items())),
            receiver_param=ListVector(list(receiver_params.items())),
        )
        actual = op.interactions.import_intercell_network(
            interactions_params=interactions_params,
            transmitter_params=transmitter_params,
            receiver_params=receiver_params,
        )

        _assert_dataframes_equal(expected, actual)



from typing import Union, ClassVar, NoReturn, Optional
from pathlib import Path
from urllib.parse import urlparse
import configparser

import attr

from omnipath.constants import License
from omnipath._core.cache._cache import Cache, FileCache, MemoryCache
from omnipath.constants._pkg_constants import DEFAULT_OPTIONS


def _is_positive(_instance, attribute: attr.Attribute, value: int) -> NoReturn:
    """Check whether the ``value`` is positive."""
    if value <= 0:
        raise ValueError(
            f"Expected `{attribute.name}` to be positive, found `{value}`."
        )


def _is_non_negative(_instance, attribute: attr.Attribute, value: int) -> NoReturn:
    """Check whether the ``value`` is non-negative."""
    if value < 0:
        raise ValueError(
            f"Expected `{attribute.name}` to be non-negative, found `{value}`."
        )


def _is_valid_url(_instance, _attribute: attr.Attribute, value: str) -> NoReturn:
    """Check whether the ``value`` forms a valid URL."""
    pr = urlparse(value)

    if not pr.scheme or not pr.netloc:
        raise ValueError(f"Invalid URL: `{value}`.")


def _cache_converter(value: Optional[Union[str, Path, Cache]]) -> Cache:
    """Convert ``value`` to :class:`omnipath._core.cache.Cache`."""
    if isinstance(value, Cache):
        return value

    if value is None:
        return MemoryCache()

    return FileCache(value)


@attr.s
class Options:
    """
    Class defining various :mod:`omnipath` options.

    Parameters
    ----------
    url
        URL of the web service.
    license
        License to use when fetching the data.
    password
        Password used when performing requests.
    cache
        Type of a cache. If `None`, cache files to memory. If a :class:`str`, persist files into a directory.
    autoload
        Whether to contact the server at ``url`` during import to get the server version and the most up-to-date
        query parameters and their valid options.
    convert_dtypes
        Whether to convert the data types of the resulting :class:`pandas.DataFrame`.
    num_retries
        Number of retries before giving up.
    timeout
        Timeout in seconds when awaiting response.
    chunk_size
        Size in bytes in which to read the data.
    progress_bar
        Whether to show the progress bar when downloading data.
    """

    config_path: ClassVar[Path] = Path.home() / ".config" / "omnipathdb.ini"

    url: str = attr.ib(
        default=DEFAULT_OPTIONS.url,
        validator=[attr.validators.instance_of(str), _is_valid_url],
        on_setattr=attr.setters.validate,
    )
    license: License = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of((str, License))),
        converter=(lambda l: None if l is None else License(l)),
        on_setattr=attr.setters.convert,
    )
    password: Optional[str] = attr.ib(
        default=None,
        repr=False,
        validator=attr.validators.optional(attr.validators.instance_of(str)),
        on_setattr=attr.setters.validate,
    )

    cache: Cache = attr.ib(
        default=DEFAULT_OPTIONS.cache_dir,
        converter=_cache_converter,
        kw_only=True,
        on_setattr=attr.setters.convert,
    )
    autoload: bool = attr.ib(
        default=DEFAULT_OPTIONS.autoload,
        validator=attr.validators.instance_of(bool),
        on_setattr=attr.setters.validate,
    )
    convert_dtypes: bool = attr.ib(
        default=DEFAULT_OPTIONS.convert_dtypes,
        validator=attr.validators.instance_of(bool),
        on_setattr=attr.setters.validate,
    )

    num_retries: int = attr.ib(
        default=DEFAULT_OPTIONS.num_retries,
        validator=[attr.validators.instance_of(int), _is_non_negative],
        on_setattr=attr.setters.validate,
    )
    timeout: Union[int, float] = attr.ib(
        default=DEFAULT_OPTIONS.timeout,
        validator=[attr.validators.instance_of((int, float)), _is_positive],
        on_setattr=attr.setters.validate,
    )
    chunk_size: int = attr.ib(
        default=DEFAULT_OPTIONS.chunk_size,
        validator=[attr.validators.instance_of(int), _is_positive],
        on_setattr=attr.setters.validate,
    )

    progress_bar: bool = attr.ib(
        default=True,
        repr=False,
        validator=attr.validators.instance_of(bool),
        on_setattr=attr.setters.validate,
    )

    def _create_config(self, section: Optional[str] = None):
        section = self.url if section is None else section
        _is_valid_url(None, None, section)
        config = configparser.ConfigParser()
        # do not save the password
        config[section] = {
            "license": str(None if self.license is None else self.license.value),
            "cache_dir": str(self.cache.path),
            "autoload": self.autoload,
            "convert_dtypes": self.convert_dtypes,
            "num_retries": self.num_retries,
            "timeout": self.timeout,
            "chunk_size": self.chunk_size,
            "progress_bar": self.progress_bar,
        }

        return config

    @classmethod
    def from_config(cls, section: Optional[str] = None) -> "Options":
        """
        Return the options from a configuration file.

        Parameters
        ----------
        section
            Section of the `.ini` file from which to create the options. It corresponds to the URL of the server.
            If `None`, use default URL.

        Returns
        -------
        :class:`omnipath._cores.utils.Options`
            The options.
        """
        if not cls.config_path.is_file():
            return cls().write()

        config = configparser.ConfigParser(default_section=DEFAULT_OPTIONS.url)
        config.read(cls.config_path)

        section = DEFAULT_OPTIONS.url if section is None else section
        _is_valid_url(None, None, section)
        _ = config.get(section, "cache_dir")

        cache = config.get(section, "cache_dir", fallback=DEFAULT_OPTIONS.cache_dir)
        cache = None if cache == "None" else cache
        license = config.get(section, "license", fallback=DEFAULT_OPTIONS.license)
        license = None if license == "None" else License(license)

        return cls(
            url=section,
            license=license,
            num_retries=config.getint(
                section, "num_retries", fallback=DEFAULT_OPTIONS.num_retries
            ),
            timeout=config.getfloat(
                section, "timeout", fallback=DEFAULT_OPTIONS.timeout
            ),
            chunk_size=config.getint(
                section, "chunk_size", fallback=DEFAULT_OPTIONS.chunk_size
            ),
            progress_bar=config.getboolean(
                section, "progress_bar", fallback=DEFAULT_OPTIONS.progress_bar
            ),
            autoload=config.getboolean(
                section, "autoload", fallback=DEFAULT_OPTIONS.autoload
            ),
            convert_dtypes=config.getboolean(
                section, "convert_dtypes", fallback=DEFAULT_OPTIONS.convert_dtypes
            ),
            cache=cache,
        )

    @classmethod
    def from_options(cls, options: "Options", **kwargs) -> "Options":
        """
        Create new options from previous options.

        Parameters
        ----------
        options
            Options from which to create new ones.
        **kwargs
            Keyword arguments overriding attributes from ``options``.

        Returns
        -------
            The newly created option.
        """
        if not isinstance(options, Options):
            raise TypeError(
                f"Expected `options` to be of type `Options`, found `{type(options)}`."
            )

        kwargs = {k: v for k, v in kwargs.items() if hasattr(options, k)}

        return cls(**{**options.__dict__, **kwargs})

    def write(self, section: Optional[str] = None) -> NoReturn:
        """Write the current options to a configuration file."""
        self.config_path.parent.mkdir(exist_ok=True)

        with open(self.config_path, "w") as fout:
            self._create_config(section).write(fout)

        return self

    def __enter__(self):
        return self.from_options(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


options = Options.from_config()


__all__ = [options, Options]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

from datetime import datetime

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
import os
import sys

sys.path.insert(0, os.path.abspath("_ext"))

import omnipath

needs_sphinx = "3.0"

# -- Project information -----------------------------------------------------

project = "omnipath"
author = omnipath.__author__
copyright = f"{datetime.now():%Y}, {author}"
github_repo = "https://github.com/saezlab/omnipath"

# The full version, including alpha/beta/rc tags
release = f"master ({omnipath.__version__})"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
    "sphinx.ext.intersphinx",
    "sphinx_paramlinks",
    "sphinx.ext.autosummary",
    "sphinx_last_updated_by_git",
    "edit_on_github",
    "enum_tools.autoenum",
]
intersphinx_mapping = dict(
    python=("https://docs.python.org/3", None),
    pandas=("https://pandas.pydata.org/pandas-docs/stable/", None),
    networkx=("https://networkx.github.io/documentation/stable/", None),
)

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]
source_suffix = [".rst"]
master_doc = "index"
pygments_style = "sphinx"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
autosummary_generate = True
autodoc_member_order = "alphabetical"
autodoc_typehints = "signature"
autodoc_docstring_signature = True
autodoc_follow_wrapped = False
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_use_rtype = True
napoleon_use_param = True
napoleon_custom_sections = [("Params", "Parameters")]
todo_include_todos = False

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_theme_options = dict(navigation_depth=4, logo_only=True)
html_show_sphinx = False

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


def setup(app):
    app.add_css_file("css/custom.css")



from abc import ABCMeta
from enum import Enum, EnumMeta
from typing import (
    Any,
    Set,
    List,
    Union,
    Mapping,
    Iterable,
    Optional,
    Sequence,
    FrozenSet,
)
from urllib.parse import urljoin
import json
import logging

from omnipath._core.utils._docs import d
from omnipath._core.query._types import (
    Int_t,
    Str_t,
    Bool_t,
    None_t,
    Strseq_t,
    License_t,
    Organism_t,
)
from omnipath._core.utils._options import Options
from omnipath.constants._constants import NoValue
from omnipath.constants._pkg_constants import Key, Format
from omnipath._core.downloader._downloader import Downloader


def _to_string_set(item: Union[Any, Sequence[Any]]) -> Set[str]:
    """
    Convert ``item`` to a `str` set.

    Parameters
    ----------
    item
        Item to convert. If it's not a sequence, it will be made into singleton.

    Returns
    -------
    :class:`set`
        Set of `str`.
    """
    if isinstance(item, (str, Enum)) or not isinstance(item, Iterable):
        item = (item,)
    return set({str(i.value if isinstance(i, Enum) else i) for i in item})


class ServerValidatorMeta(EnumMeta, ABCMeta):  # noqa: D101
    class Validator:
        """
        Class that validates values for some parameter passed to the server.

        Parameters
        ----------
        param
            Name of the parameter we're checking. Only used for informing the user.
        haystack
            Valid values for the ``paramter``. If `None`, no validation will be performed.
        doc
            Doctring specific to the ``param``.
        """

        def __init__(
            self,
            param: str,
            haystack: Optional[Set[str]] = None,
            doc: Optional[str] = None,
        ):
            if isinstance(haystack, str):
                haystack = (haystack,)
            elif haystack is not None and not isinstance(haystack, Iterable):
                raise TypeError(
                    f"Expected `haystack` for `{param}` to be either a "
                    f"`str` or a  `Sequence`, found `{type(haystack)}`."
                )

            self._param = param.lower()
            self._haystack = haystack if haystack is None else frozenset(haystack)
            self._query_doc_ = None if not doc else doc  # doc can also be `()`

        @property
        def haystack(self) -> Optional[FrozenSet[str]]:
            """Return the valid values for this parameter."""
            return self._haystack

        def __call__(self, needle: Optional[Set[str]]) -> Optional[Set[str]]:
            """
            Check whether ``needle`` is a valid value for :paramref:`_param`.

            Parameters
            ----------
            needle
                Needle to check.

            Returns
            -------
                `None` if the ``needle`` was `None`, otherwise the ``needle`` as a `str` set,
                optionally intersected with :paramref:`_haystack` if it is not `None`.

            Raises
            ------
            ValueError
                If :paramref:`_haystack` is not `None` and no valid values were found.
            """
            if needle is None:
                return None
            elif isinstance(needle, bool):
                needle = int(needle)
            elif isinstance(needle, Enum):
                needle = needle.value

            needle = _to_string_set(needle)
            if self.haystack is None:
                logging.debug(
                    f"Unable to perform parameter validation for `{self._param}`, haystack is empty"
                )
                return needle

            res = needle & self.haystack
            if not len(res):
                raise ValueError(
                    f"No valid options found for parameter `{self._param}` in: `{sorted(needle)}`.\n"
                    f"Valid options are: `{sorted(self.haystack)}`."
                )
            elif len(res) < len(needle):
                logging.warning(
                    f"Encountered invalid value(s) for `{self._param}`. "
                    f"Remaining values are `{sorted(res)}`"
                )

            return res

    def __new__(cls, clsname, superclasses, attributedict):  # noqa: D102
        from omnipath import options

        endpoint = attributedict.pop(
            "__endpoint__", clsname.lower().replace("validator", "")
        )
        use_default = True
        old_members = list(attributedict._member_names)
        old_values = cls._remove_old_members(attributedict)

        if endpoint is None:
            if len(old_members):
                raise ValueError(
                    "If `__endpoint__` is `None`, no members must be specified."
                )
        elif options.autoload:
            use_default = False
            with Options.from_options(
                options,
                num_retries=0,
                timeout=0.1,
                cache=None,
                progress_bar=False,
                chunk_size=2048,
            ) as opt:
                try:
                    logging.debug("Attempting to construct classes from the server")
                    res = Downloader(opt).maybe_download(
                        urljoin(urljoin(opt.url, f"{Key.QUERIES.s}/"), endpoint),
                        callback=json.load,
                        params={Key.FORMAT.s: Format.JSON.s},
                    )

                    if len({str(k).upper() for k in res.keys()}) != len(res):
                        raise RuntimeError(
                            f"After upper casing, key will not be unique: `{list(res.keys())}`."
                        )

                    for k, value in res.items():
                        if (
                            isinstance(value, str)
                            and "no such query available" in value
                        ):
                            raise RuntimeError(f"Invalid endpoint: `{endpoint}`.")

                        key = str(k).upper()
                        if value is None:
                            attributedict[key] = cls.Validator(param=k)
                        elif isinstance(value, Sequence):
                            attributedict[key] = cls.Validator(
                                param=k, haystack={str(v) for v in value}
                            )
                        else:
                            attributedict[key] = cls.Validator(param=k)
                except Exception as e:
                    logging.debug(
                        f"Unable to construct classes from the server. Reason: `{e}`"
                    )
                    use_default = True

        if use_default:
            if endpoint is not None:
                logging.debug(
                    f"Using predefined class: `{clsname}`."
                    + (
                        ""
                        if options.autoload
                        else " Consider specifying `omnipath.options.autoload = True`"
                    )
                )

            _ = cls._remove_old_members(attributedict)
            for k, v in zip(old_members, old_values):
                attributedict[k] = cls.Validator(param=k, doc=v)

        return super().__new__(cls, clsname, superclasses, attributedict)

    @classmethod
    def _remove_old_members(cls, attributedict) -> List[Any]:
        vals = []
        for k in list(attributedict._member_names):
            vals.append(attributedict.pop(k, None))
        attributedict._member_names = []

        return vals


class AutoValidator(NoValue):  # noqa: D101
    @property
    def valid(self) -> Optional[Set[str]]:
        """Return the valid values."""
        return self.value.haystack

    @property
    def annotation(self) -> Mapping[str, type]:
        """Return the type annotations."""
        return getattr(self, "__annotations__", {}).get(self.name, Any)

    @property
    def doc(self) -> Optional[str]:
        """Return the docstring."""
        return getattr(self.value, "_query_doc_", None)

    @d.dedent
    def __call__(self, value: Union[str, Sequence[str]]) -> Optional[Set[str]]:
        """%(validate)s"""  # noqa: D401
        return self.value(value)


class QueryValidatorMixin(AutoValidator, metaclass=ServerValidatorMeta):  # noqa: D101
    __endpoint__ = None


class EnzsubValidator(QueryValidatorMixin):  # noqa: D101
    DATABASES: Strseq_t = ()
    ENZYME_SUBSTRATE: Str_t = ()
    ENZYMES: Strseq_t = ()
    FIELDS: Strseq_t = ()
    FORMAT: Str_t = ()
    GENESYMBOLS: Bool_t = ()
    HEADER: Str_t = ()
    LICENSE: License_t = ()
    LIMIT: Int_t = ()
    MODIFICATION: Str_t = ()
    ORGANISMS: Organism_t = ()
    PARTNERS: Strseq_t = ()
    PASSWORD: Str_t = ()
    RESIDUES: Strseq_t = ()
    RESOURCES: Strseq_t = ()
    SUBSTRATES: Strseq_t = ()
    TYPES: Strseq_t = ()


class InteractionsValidator(QueryValidatorMixin):  # noqa: D101
    DATABASES: Strseq_t = ()
    DATASETS: Strseq_t = ()
    DIRECTED: Bool_t = ()
    DOROTHEA_LEVELS: Strseq_t = ()
    DOROTHEA_METHODS: Strseq_t = ()
    ENTITY_TYPES: Strseq_t = ()
    FIELDS: Strseq_t = ()
    FORMAT: Str_t = ()
    GENESYMBOLS: Bool_t = ()
    HEADER: Str_t = ()
    LICENSE: License_t = ()
    LIMIT: Int_t = ()
    ORGANISMS: Organism_t = ()
    PARTNERS: Strseq_t = ()
    PASSWORD: Str_t = ()
    RESOURCES: Strseq_t = ()
    SIGNED: Bool_t = ()
    SOURCE_TARGET: Bool_t = ()
    SOURCES: Strseq_t = ()
    TARGETS: Strseq_t = ()
    TFREGULONS_LEVELS: Strseq_t = ()
    TFREGULONS_METHODS: Strseq_t = ()
    TYPES: Strseq_t = ()


class ComplexesValidator(QueryValidatorMixin):  # noqa: D101
    DATABASES: Strseq_t = ()
    FIELDS: Strseq_t = ()
    FORMAT: Str_t = ()
    HEADER: Str_t = ()
    LICENSE: License_t = ()
    LIMIT: Int_t = ()
    PASSWORD: Str_t = ()
    PROTEINS: Strseq_t = ()
    RESOURCES: Strseq_t = ()


class AnnotationsValidator(QueryValidatorMixin):  # noqa: D101
    DATABASES: Strseq_t = ()
    ENTITY_TYPES: Strseq_t = ()
    FIELDS: Strseq_t = ()
    FORMAT: Str_t = ()
    GENESYMBOLS: Bool_t = ()
    HEADER: Str_t = ()
    LICENSE: License_t = ()
    LIMIT: Int_t = ()
    PASSWORD: Str_t = ()
    PROTEINS: Strseq_t = ()
    RESOURCES: Strseq_t = ()


class IntercellValidator(QueryValidatorMixin):  # noqa: D101
    ASPECT: Str_t = ()
    CATEGORIES: Str_t = ()
    CAUSALITY: Str_t = ()
    DATABASES: Strseq_t = ()
    ENTITY_TYPES: Str_t = ()
    FIELDS: Strseq_t = ()
    FORMAT: Str_t = ()
    HEADER: None_t = ()
    LICENSE: License_t = ()
    LIMIT: Int_t = ()
    PARENT: Str_t = ()
    PASSWORD: Str_t = ()
    PLASMA_MEMBRANE_PERIPHERAL: Bool_t = ()
    PLASMA_MEMBRANE_TRANSMEMBRANE: Bool_t = ()
    PMP: Bool_t = ()
    PMTM: Bool_t = ()
    PROTEINS: Strseq_t = ()
    REC: Bool_t = ()
    RECEIVER: Strseq_t = ()
    RESOURCES: Strseq_t = ()
    SCOPE: Str_t = ()
    SEC: Bool_t = ()
    SECRETED: Bool_t = ()
    SOURCE: Str_t = ()
    TOPOLOGY: Str_t = ()
    TRANS: Bool_t = ()
    TRANSMITTER: Bool_t = ()


__all__ = [
    EnzsubValidator,
    InteractionsValidator,
    ComplexesValidator,
    AnnotationsValidator,
    IntercellValidator,
]



from typing import Any, Union, Mapping, Iterable, Optional
import logging

import pandas as pd

from omnipath._core.query import QueryType
from omnipath._core.requests._request import OrganismGenesymbolsRemover
from omnipath.constants._pkg_constants import final


@final
class Complexes(OrganismGenesymbolsRemover):
    """Request information about protein complexes from [OmniPath]_."""

    __string__ = frozenset(
        {
            "name",
            "components",
            "components_genesymbols",
            "stoichiometry",
            "references",
            "identifiers",
        }
    )
    __categorical__ = frozenset({"sources"})

    _query_type = QueryType.COMPLEXES

    def _resource_filter(self, data: Mapping[str, Any], **_) -> bool:
        return True

    @classmethod
    def complex_genes(
        cls,
        genes: Union[str, Iterable[str]],
        complexes: Optional[pd.DataFrame] = None,
        total_match: bool = False,
    ) -> pd.DataFrame:
        """
        Get all the molecular complexes for a given ``genes``.

        This function returns all the molecular complexes where an input set of genes participate. User can choose
        to retrieve every complex where any of the input genes participate or just retrieve these complexes where
        all the genes in input set participate together.

        Parameters
        ----------
        genes
            The genes for which complexes will be retrieved (hgnc format).
        complexes
            Complex data from :meth:`get`. If `None`, new request will be made.
        total_match
            If `True`, get only complexes where all the genes participate together, otherwise get complexes
            where any of the genes participate.

        Returns
        -------
        :class:`pandas.DataFrame`
            The filtered ``complexes``.
        """
        if isinstance(genes, str):
            genes = (genes,)
        genes = tuple(set(genes))
        if not len(genes):
            raise ValueError("No genes have been selected.")

        if complexes is None:
            logging.info("Fetching complexes from the server")
            complexes = cls.get()
        if not isinstance(complexes, pd.DataFrame):
            raise TypeError(
                f"Expected `complexes` to be of type `pandas.DataFrame`, found `{type(complexes)}`."
            )

        if complexes.empty:
            logging.warning("Complexes are empty")
            return complexes

        col = "components_genesymbols"
        if col not in complexes:
            raise KeyError(f"Unable to find `{col}` in `{complexes.columns}`.")

        reduction = all if total_match else any

        return complexes.loc[
            complexes[col]
            .str.split("_")
            .apply(lambda needles: reduction(n in genes for n in needles))
        ].reset_index(drop=True)


__all__ = [Complexes]



#  This file is part of the `omnipath` python module
#
#  Copyright
#  2020
#  Heidelberg University
#
#  File author(s): Dénes Türei (turei.denes@gmail.com)
#
#  Distributed under the MIT (expat) License.
#  See accompanying file LICENSE or copy at
#      https://opensource.org/licenses/MIT
#
#  Website: http://omnipathdb.org/
#
from pathlib import Path

from setuptools import setup, find_packages

try:
    from omnipath import __email__, __author__, __version__, __maintainer__
except ImportError:
    __author__ = "Michal Klein, Dénes Türei"
    __maintainer__ = "Michal Klein, Dénes Türei"
    __email__ = "turei.denes@gmail.com"
    __version__ = "1.0.4"


setup(
    # general
    name="omnipath",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    # version
    version=__version__,
    author=__author__,
    author_email=__email__,
    maintainer=__maintainer__,
    maintainer_email=__email__,
    # description
    description=Path("README.rst").read_text("utf-8").splitlines()[2],
    long_description=Path("README.rst").read_text("utf-8"),
    long_description_content_type="text/x-rst; charset=UTF-8",
    # links
    url="https://omnipathdb.org/",
    download_url="https://github.com/saezlab/omnipath/releases/",
    project_urls={
        "Documentation": "https://omnipath.readthedocs.io/en/latest",
        "Source Code": "https://github.com/saezlab/omnipath",
    },
    license="MIT",
    platforms=["Linux", "Unix", "MacOSX", "Windows"],
    # keywords
    keywords=sorted(
        {
            "protein",
            "mRNA",
            "miRNA",
            "DNA",
            "signaling",
            "SignaLink",
            "SIGNOR",
            "InnateDB",
            "IntAct",
            "Reactome",
            "MPPI",
            "NCI-PID",
            "DIP",
            "MatrixDB",
            "PANTHER",
            "PhosphoSite",
            "PhosphoPoint",
            "DEPOD",
            "SPIKE",
            "KEGG",
            "Autophagy",
            "ARN",
            "NRF2ome",
            "Guide to Pharmacology",
            "UniProt",
            "BioPAX",
            "Ensembl",
            "Surfaceome",
            "Exocarta",
            "Vesiclepedia",
            "Matrisome",
            "Human Protein Atlas",
            "Compleat",
            "CORUM",
            "ComplexPortal",
            "BioGRID",
            "STRING",
            "ICELLNET",
            "Cell Surface Protein Atlas",
            "COSMIC",
            "Cancer Gene Census",
            "IntOGen",
            "TopDB",
            "iTALK",
            "Human Plasma Membrane Receptome",
            "EMBRACE",
            "ELM",
            "phospho.ELM",
            "CancerSEA",
            "ComPPI",
            "CellPhoneDB",
            "DGIdb",
            "DisGeNet",
            "PAZAR",
            "ORegAnno",
            "TRED",
            "DoRothEA",
            "TRRD",
            "CPAD",
            "regulation",
            "phosphorylation",
            "kinase",
            "phosphatase",
            "dephosphorylation",
            "directed graph",
            "annotations",
            "cancer",
            "complexes",
            "intercellular communication",
            "HGNC",
            "GPCRdb",
            "MSigDB",
            "GSEA",
            "Phobius",
            "Phosphatome",
            "NetPath",
            "gene",
            "gene symbol",
            "mouse",
            "rat",
            "HomoloGene",
            "integrin",
            "adhesion",
            "receptor",
            "ligand",
            "transporter",
            "ion channel",
            "disease",
            "activity flow",
            "transcription",
            "PPI",
            "subcellular localization",
            "pathway",
            "signaling pathway",
        }
    ),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Typing :: Typed",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    # package installation
    packages=find_packages(),
    zip_safe=False,
    python_requires=">=3.7",
    include_package_data=False,
    # dependency_links = deplinks
    install_requires=list(
        map(
            str.strip,
            Path("requirements.txt").read_text("utf-8").splitlines(),
        )
    ),
    extras_require={
        "graph": ["networkx>=4.3.0"],
        "tests": ["tox>=3.20.1"],
        "docs": [
            line
            for line in (Path("docs") / "requirements.txt")
            .read_text("utf-8")
            .splitlines()
            if not line.startswith("-r")
        ],
        "dev": ["pre-commit>=2.7.1", "bump2version"],
    },
)



from docrep import DocstringProcessor

_general_get = """
Perform the query.

Parameters
----------
kwargs
    Parameters of the request. For more information about available values, see :meth:`params`.

Returns
-------
:class:`pandas.DataFrame`
    The result of this query."""
_interactions_datasets = """
    - :attr:`omnipath.constants.InteractionDataset.OMNIPATH`
    - :attr:`omnipath.constants.InteractionDataset.PATHWAY_EXTRA`
    - :attr:`omnipath.constants.InteractionDataset.KINASE_EXTRA`
    - :attr:`omnipath.constants.InteractionDataset.LIGREC_EXTRA`
    - :attr:`omnipath.constants.InteractionDataset.DOROTHEA`
    - :attr:`omnipath.constants.InteractionDataset.TF_TARGET`
    - :attr:`omnipath.constants.InteractionDataset.TF_MIRNA`
    - :attr:`omnipath.constants.InteractionDataset.TF_REGULONS`
    - :attr:`omnipath.constants.InteractionDataset.MIRNA_TARGET`
    - :attr:`omnipath.constants.InteractionDataset.LNCRNA_MRNA`"""
_validate = """
Validate the ``value`` for the :paramref:`param`.

Parameters
----------
value
    Value to validate.

Returns
-------
    The valid values."""
_query_resources = """
    Return the available resources for this query."""
_query_params = """
    Return the available values for each parameter, if available."""

d = DocstringProcessor(
    general_get=_general_get,
    interaction_datasets=_interactions_datasets,
    validate=_validate,
    query_params=_query_params,
    query_resources=_query_resources,
)
